
from .read import *
from .read_general import *
from .utilities import *

__all__ = [
    'loadtxt',
    'read_dcd',
    'read_peptide_sequence',
    'read_pdb',
    'read_psf',
    'read_csv',
    'read_table',
    'vglob'
]

# TODO write IC coordinates for CHARMM



"""
Boltzmann
"""

import numpy as np

from scipy.constants import Boltzmann, Avogadro, atmosphere
import scipy.constants

__all__ = [
    'AVOGADRO',
    'BOLTZMANN'
]


# Unit class
class Unit:
    def __init__(self, symbol):
        self.magnitude = ''
        self.symbol = symbol
        self.inverse = False

    def __repr__(self):
        return self.to_string()

    def __truediv__(self, other):
        u = ComplexUnit()
        u.add(self)
        u.add(other.invert())
        return u

    def copy(self):
        u = Unit(symbol=self.symbol)
        u.magnitude = self.magnitude
        u.inverse = self.inverse
        return u

    def invert(self):
        u = self.copy()
        u.inverse = True
        return u

    def to_string(self, ignore_inverse=True):
        repr = self.magnitude + self.symbol
        if not ignore_inverse and self.inverse:
            repr = f'1 / {repr}'
        return repr


class ComplexUnit:
    def __init__(self):
        self._units = []

    def __repr__(self):
        num = []
        den = []
        for unit in self._units:
            repr = unit.to_string(ignore_inverse=True)
            if unit.inverse:
                den.append(repr)
            else:
                num.append(repr)
        num_str = '*'.join(num)
        den_str = '*'.join(den)
        if len(num) > 1:
            num_str = f'({num_str})'
        if len(den) > 1:
            den_str = f'({den_str})'
        if len(num) == 0:
            num_str = '1'
        if len(den) > 0:
            return f'{num_str}/{den_str}'
        else:
            return num_str

    def add(self, unit):
        self._units.append(unit)


# Define base units
ANGSTROM = Unit(symbol='Å')
SECOND = Unit(symbol='s')

# Define



AVOGADRO = 6.02214076e23  # scipy.constants.Avogadro
BOLTZMANN = 1.380649e-23  # scipy.constants.Boltzmann
CAL_TO_JOULE = 4.184  # 1 calorie = 4.184 joules
KCAL_TO_JOULE = CAL_TO_JOULE * 1000.


# TODO make this easily usable.
class GasConstant:
    def __init__(self):
        pass

    @property
    def kcal_per_K_mol(self):
        return per_mole(joule_to_kcal(BOLTZMANN))


# Joule to kcal conversion
def joule_to_kcal(x):
    """
    Convert joule to kcal. 1 J = 1 / 4184. kcal

    Parameters
    ----------
    x : float
        Joule.

    Returns
    -------
    float
        kcal.
    """

    return x / 4184.


# kcal to Joule conversion
def kcal_to_joule(x):
    """
    Convert kcal to joule. 1 kcal = 4184 J.

    Parameters
    ----------
    x : float
        kcal.

    Returnsx
    -------
    float
        Joule.
    """

    return 4184. * x


def per_mole(x):
    """
    Convert quantity to per mole. Divide by NA.

    Parameters
    ----------
    x : float
        Quantity.

    Returns
    -------
    float
        Quantity per mole.
    """

    return x * AVOGADRO


class Quantity:
    def __init__(self, value, units=None):
        self._value = value
        if units is None:
            units = Unit
        self._units = units

    def __repr__(self):
        return f'{self._value} {self._units.symbol}'


# Boltzmann constant in kcal/K/mol
BOLTZMANN_KCAL_K_MOL = per_mole(joule_to_kcal(BOLTZMANN))
BOLTZMANN_KCAL_MOL_K = BOLTZMANN_KCAL_K_MOL
np.testing.assert_almost_equal(BOLTZMANN * AVOGADRO / 4184., BOLTZMANN_KCAL_K_MOL)

# Pressure
pressure = Avogadro * atmosphere / 1e30 / 4184.  # 1e-30 m^3 per A^3, 4184 J per kcal

if __name__ == '__main__':
    x = Quantity(100, units=ANGSTROM)
    print(f'kb={BOLTZMANN_KCAL_MOL_K} kcal/mol')




from io import StringIO
import numpy as np
import pandas as pd


class PSF:  # this should probably be a child of Topology
    def __init__(self, atoms, bonds):
        self._atoms = atoms
        self._bonds = bonds

    def __repr__(self):
        return self._atoms.to_markdown()

    def _repr_html_(self):
        return self._atoms.to_html(max_rows=10)


def read_psf(fname):
    # Sections
    sections = {}
    counts = {}

    # Read in entire file first
    with open(fname, 'r') as buf:
        section = None
        n_remaining = 0

        # Read file taking out the good bits
        for line in buf:
            # Read section
            if section is not None and n_remaining > 0:
                sections[section] += line
                if section in ['!NTITLE', '!NATOM']:
                    n_remaining -= 1
                elif section == '!NBOND':  # at most 4 bonds per line
                    n_remaining -= 4

                # If we hit 0 (or fewer), we are done
                if n_remaining <= 0:
                    section = None

                    # Otherwise, figure out if a section is starting
            if section is None:
                data = line.strip().split()
                if len(data) > 1 and data[1] in ['!NTITLE', '!NATOM', '!NBOND:']:
                    section = data[1].replace(':', '')
                    sections[section] = ''
                    counts[section] = int(data[0])
                    n_remaining = int(data[0])

                    # Process
    atoms = pd.read_table(
        StringIO(sections['!NATOM']),
        sep='\s+',
        header=None,
        names=[
            'atom_id',
            'segment',
            'residue_id',
            'residue',
            'atom',
            'atom_type',
            'charge',
            'mass',
            'unused'
        ]
    )
    atoms.index.name = 'index'

    bonds = pd.DataFrame(
        np.array(np.array_split(sections['!NBOND'].split(), counts['!NBOND']), dtype='int'),
        columns=[
            'atom_id0',
            'atom_id1'
        ]
    )

    # Sanity
    assert len(atoms) == counts['!NATOM']
    assert len(bonds) == counts['!NBOND']

    # Return
    return PSF(
        atoms=atoms,
        bonds=bonds
    )



from .block import *
from .statistics import *
from .time_series import *

__all__ = [
    'acorr',
    'acov',
    'Block',
    'block_average',
    'block_error',
    'normalize',
    'sem',
    'sem_block',
    'sem_tcorr',
    'standardize',
    'tcorr',
]


# TODO find me a home
import numpy as np
import pandas as pd
def normalize(a):
    if isinstance(a, (pd.DataFrame, pd.Series)):
        a = a.div(a.abs().sum(axis=0), axis=1)  # noqa

    elif isinstance(a, (list, np.ndarray)):
        a = a / np.sum(np.abs(a))

    else:
        assert AttributeError('cannot normalize')

    return a


def standardize(a):
    raise NotImplementedError



import numpy as np
from scipy.spatial.distance import pdist, squareform


class Lattice2D:
    # PBC enabled
    # bonds are sequential
    def __init__(self, n_beads):
        self._n_beads = n_beads

    # This isn't right...
    def permutations(self):
        """

        3 - 2
        4 - 5
        5 - 11?

        12345
        *****

        1234*
        ***5*

        123**
        **45*

        123**
        **4**
        **5**

        123**
        *54**

        12***
        *345*

        12***
        *3***
        *45**

        12***
        *3***
        *4***
        *5***

        12***
        *3***
        54***

        12***
        43***
        5****

        *12**
        453**

        541**
         32

        Returns
        -------

        """
        n_beads = self._n_beads
        from itertools import combinations_with_replacement
        xy = list(combinations_with_replacement(range(n_beads), r=2))
        for _ in combinations_with_replacement(xy, r=n_beads):
            if all(np.diag(squareform(pdist(_)), k=1) == 1):
                yield _

    def draw_permutations(self):
        n_beads = self._n_beads
        permutations = self.permutations()
        for permutation in permutations:
            a = np.empty((n_beads, n_beads), dtype='str')
            a[:] = ' '
            for x, y in permutation:
                a[x, y] = '*'
            yield a



"""
xx1xx
xx23x
xxx4x
"""


if __name__ == '__main__':
    lattice = Lattice2D(4)
    p = lattice.draw_permutations()







"""
sequence.py

>>> abeta = Protein('YEVHHQKLVFFAEDVGSNKGAIIGLMVGGVV')
"""

import pandas as pd
from typelike import ArrayLike


class Residue:
    def __init__(self):
        pass


class AminoAcid(Residue):
    _defaults = pd.DataFrame([
        ['ALA', 'A', 0],
        ['CYS', 'C', 0],
        ['ASP', 'D', -1],
        ['GLU', 'E', -1],
        ['PHE', 'F', 0],
        ['GLY', 'G', 0],
        ['HIS', 'H', 0],
        ['ILE', 'I', 0],
        ['LYS', 'K', 1],
        ['LEU', 'L', 0],
        ['MET', 'M', 0],
        ['ASN', 'N', 0],
        ['PRO', 'P', 0],
        ['GLN', 'Q', 0],
        ['ARG', 'R', 1],
        ['SER', 'S', 0],
        ['THR', 'T', 0],
        ['VAL', 'V', 0],
        ['TRP', 'W', 0],
        ['TYR', 'Y', 0],
    ], columns=['name', 'code', 'charge'])

    def __init__(self, name, code, charge):
        super().__init__()
        self._name = name
        self._code = code
        self._charge = charge

    @property
    def name(self):
        return self._name

    @property
    def code(self):
        return self._code


def _get_amino_acid(name=None, code=None):
    if name is None and code is None:
        raise AttributeError('must specify name or code')

    name = name.upper()
    code = code.upper()

    _code_to_name = {
        'A': 'ALA',
        'C': 'CYS',
        'D': 'ASP',
        'E': 'GLU',
        'F': 'PHE',
        'G': 'GLY',
        'H': 'HIS',
        'I': 'ILE',
        'K': 'LYS',
        'L': 'LEu',
        'M': 'MET',
        'N': 'ASN',
        'P': 'PRO',
        'Q': 'GLN',
        'R': 'ARG',
        'S': 'SER',
        'T': 'THR',
        'V': 'VAL',
        'W': 'TRP',
        'Y': 'TYR'
    }

    _name_to_code = {name: code for code, name in _code_to_name.items()}

    if code is not None:
        name = _code_to_name[code]
    else:
        code = _name_to_code[name]

    return AminoAcid(name, code, 0.)


class Sequence:
    def __init__(self):
        pass


class Protein(Sequence):
    """
    Construct for dealing with protein sequences.
    """

    def __init__(self, sequence, ):
        """
        Initialize Protein object.

        Parameters
        ----------
        sequence : np.ndarray
            List of residues.
        """

        # Convert to list if not already list
        # super().__init__()
        # if not isinstance(sequence, ArrayLike):
        if isinstance(sequence, str):
            sequence = list(sequence)

        # Which format are these residues in?
        len_sequence = len(sequence[0])
        if len_sequence == 1:
            sequence = _letter_to_code(sequence)
        elif len_sequence != 3:
            raise AttributeError('must supply residue triplets')

        # Save
        self.sequence = sequence

    def __repr__(self):
        return str(self.sequence)

    def to_letters(self, join=False):
        code_to_letter = {
            'ALA': 'A',
            'ARG': 'R',
            'ASN': 'N',
            'ASP': 'D',
            'CYS': 'C',
            'GLN': 'Q',
            'GLU': 'E',
            'GLY': 'G',
            'HIS': 'H',
            'HSD': 'H',
            'ILE': 'I',
            'LEU': 'L',
            'LYS': 'K',
            'MET': 'M',
            'PHE': 'F',
            'PRO': 'P',
            'SER': 'S',
            'THR': 'T',
            'TRP': 'W',
            'TYR': 'Y',
            'VAL': 'V',
        }

        sequence = [code_to_letter[code] for code in self.sequence]

        if join:
            sequence = ''.join(sequence)

        return sequence

    def to_str(self):
        pass


def _letter_to_code(residues):
    # Make sure we're in the right format
    # noinspection DuplicatedCode
    letter_to_code = {
        'A': 'ALA',
        'R': 'ARG',
        'N': 'ASN',
        'D': 'ASP',
        'C': 'CYS',
        'Q': 'GLN',
        'E': 'GLU',
        'G': 'GLY',
        'H': 'HIS',
        'I': 'ILE',
        'L': 'LEU',
        'K': 'LYS',
        'M': 'MET',
        'F': 'PHE',
        'P': 'PRO',
        'S': 'SER',
        'T': 'THR',
        'W': 'TRP',
        'Y': 'TYR',
        'V': 'VAL'
    }
    return [letter_to_code.get(residue, residue) for residue in residues]


if __name__ == '__main__':
    ab = Protein('YEVHHQKLVFFAEDVGSNKGAIIGLMVGGVV')  # noqa




from .bonds import *
from .distance import *
from .histogram import *
from .rmsd import *
from .sanity import *

from . import protein
from .protein import *

__all__ = [
    'angle',
    'contacts',
    'contacts_to_vector',
    'distance',
    'distances',
    'has_cross_interactions',
    'ihist',
    'rmsd',
]

__all__.extend(protein.__all__)



"""
replica_exchange.py

author: C. Lockhart <clockha2@gmu.edu>
language: Python3
"""

from iox import loadtxt

import numpy as np
import pandas as pd


# Class that stores exchange history, i.e., which replicas visited which states over time
class ExchangeHistory:
    """
    Replica-exchange follows replicas over time as they walk across configuration space. Here, "configuration" is used
    generically and can refer to temperature (commonly), Hamiltonian scaling factors, etc. The "configuration" refers
    to all set parameters.

    The purpose of this class is to permit easy evaluation of exchange performance. We can compute (and plot) parameters
    like the exchange rate, mosaic plot of the replica walk, the replica-mixing parameter, and tunneling time.
    """

    __slots__ = '_data'

    # Initialize instance of exchange history
    # TODO allow "warm start" from initial conditions
    def __init__(self, data, only_neighbors=True):
        """
        `data` must be a pandas DataFrame with the columns "replica", "config", and "step". Note that the elements in
        "replica" and "config" refer to indices.

        Parameters
        ----------
        data : pandas.DataFrame
            Exchange history.
        only_neighbors : bool
            Were exchanges permitted only between neighbors?
        """

        # Validate input
        if not isinstance(data, pd.DataFrame) or ~np.in1d(['replica', 'config', 'step'], data.columns).all():
            raise AttributeError('data must be DataFrame')

        # Make sure that replicas do not jump by more than 1 configuration at a time if only_neighbors is true
        # TODO is it necessary to have this validation step?
        if only_neighbors:
            n_replicas = data['replica'].nunique()
            data = data.sort_values(['replica', 'step'])
            for replica in range(n_replicas):
                tmp = data.query(f'replica == {replica}')['config'].diff().fillna(0)
                assert tmp.min() == -1
                assert tmp.max() == 1

        # Save
        self._data = data

    # Read NAMD history file (from their RE multi-copy algorithm)
    # TODO what if someone loaded .sort.history? The labels replica and config would be swapped.
    @classmethod
    def from_namd(cls, fname, n_replicas, glob=False):
        """
        Construct the exchange history from the results of NAMD replica exchange.

        Parameters
        ----------
        fname : str
            Name of the history file. The assumption is that there is one history file per replica. This method expects
            to find the variable `{replica}` to specify the file for each replica.
        n_replicas : int
            The number of replicas.
        glob : bool
            Indicates if `fname` contains extra glob-like features, such as wild-cards.

        Returns
        -------
        ExchangeHistory
            Newly-created instance of ExchangeHistory.

        Examples
        --------
        >>> ExchangeHistory.from_namd('exchange_{replica}.history', n_replicas=16, glob=False)
        >>> ExchangeHistory.from_namd('exchange_job?.{replica}.history', n_replicas=16, glob=True)
        """

        # Initialize DataFrame to hold results
        data = pd.DataFrame()

        # Loop over all replicas and read in files
        for replica in range(n_replicas):
            tmp = loadtxt(fname.format(replica=replica), glob=glob)
            data = data.append(pd.DataFrame({
                'step': tmp[:, 0],
                'replica': np.repeat(replica, len(tmp)),
                'config': tmp[:, 1].astype(int),
                'temperature': tmp[:, 2]
            }), ignore_index=True)

        # Return instantiated ExchangeHistory class. This is sorted but there's strictly no reason to do this.
        return cls(data.sort_values(['replica', 'step']))

    # Read history from parquet
    @classmethod
    def from_parquet(cls, fname):
        """
        Read exchange history from parquet.

        Parameters
        ----------
        fname : str
            Parquet file.

        Returns
        -------
        ExchangeHistory
        """

        return cls(pd.read_parquet(fname))

    @property
    def n_configs(self):
        return self._data['config'].nunique()

    @property
    def n_replicas(self):
        return self._data['replica'].nunique()

    @property
    def n_steps(self):
        return self._data['step'].nunique()

    # Cross-tabulate by config and replica axes
    def crosstab(self, index='config', column='replica'):
        """
        Cross-tabulate by "config" and "replica" axes. In other words, count the number of steps at given config and
        replica indices.

        Parameters
        ----------
        index : str
            Axis to define rows.
        column : str
            Axis to define columns.

        Returns
        -------
        pandas.DataFrame
            Cross-tabulation of the ExchangeHistory instance.
        """

        # Get the exchange history trajectory by the index
        data = self.trajectory(by=index, reset_index=True)

        # Melt the data by the column
        data_melt = data.melt(value_name=column)

        # Return pandas cross tabulation
        return pd.crosstab(index=data_melt[index], columns=data_melt[column])

    # Compute the exchange rate
    def exchange_rate(self, by='config', add_initial_state=True, n_attempts=None, terminal_factor=0.5):
        """
        Compute the exchange rate.

        Parameters
        ----------
        by : str
            Default: config
        add_initial_state : bool
        n_attempts : int
            Number of attempts per config or replica.
        terminal_factor : float
            Factor to apply to the first and last config to reduce its number of attempts.

        Returns
        -------
        pandas.Series
            Exchange rate per config or replica.
        """

        # Pull the trajectory
        trj = self.trajectory(by=by if by is not None else 'config')
        
        #
        if add_initial_state:
            initial_state = pd.DataFrame(
                [np.arange(len(trj.columns), dtype='int')],
                columns=trj.columns
            )
            trj = pd.concat([initial_state, trj])

        # Derive number of attempts. This is tricky because it depends on how exchange was set up. In unbiased exchange,
        # we expect that every replica/config exchanges every step. However, the terminal configs usually only exchange
        # every other attempt (because they're at a boundary) so we apply the terminal factor. In the future, I would
        # like to make this more intuitive.
        if n_attempts is None:
            n_attempts = np.repeat(len(trj), len(trj.columns))
            n_attempts[0] = np.floor(n_attempts[0] * terminal_factor)
            n_attempts[-1] = np.floor(n_attempts[-1] * terminal_factor)

        # Compute number of exchanges. The logic here is that if the index from one step to the next changes, then an
        # exchange has happened.
        n_exchanges = (trj.diff().dropna(axis=0) != 0).sum()

        # Return the rate
        return np.sum(n_exchanges) / np.sum(n_attempts) if by is None else n_exchanges / n_attempts

    # TODO incorporate this into above
    def direction_exchange_rate(self):
        trj = self.trajectory(by='config')  # must be by the non-walker
        n_steps, n_states = trj.shape

        n_even_steps = (np.mod(trj.index.to_numpy(), 2) == 0).sum()
        n_odd_steps = n_steps - n_even_steps

        df = pd.DataFrame(columns=['state', 'direction', 'n_exchanges', 'n_attempts']).set_index(['state', 'direction'])
        for state in range(n_states):
            for direction in [-1, 1]:
                if (state == 0 and direction == -1) or (state == n_states - 1 and direction == 1):
                    continue
                df.loc[(state, direction), 'n_exchanges'] = (trj[state] == trj[state + direction].shift(-1)).sum()
                df.loc[(state, direction), 'n_attempts'] = n_even_steps if direction == 1 else n_odd_steps
        return df

    # TODO change to mixing_parameter
    def hansmann(self, ):
        r"""
        The Hansmann parameter :math:`h(T)` shows the residence time :math:`\tau` replica :math:`r` (of :math:`R` total
        replicas) spends at configuration :math:`T`.

        .. math:: h(T) = 1 - \frac{\sqrt{\sum_{r=1}^R \tau_r^2}}{\sum_{r=1}^R \tau_r}

        If all replicas are equally sampled across all configurations, then :math:`h(T) = 1 - 1 / \sqrt{R}`.

        Returns
        -------
        pandas.Series
            Hansmann parameter computed for all configurations.
        """

        # Cross-tabulate replica by configuration
        data = self.crosstab(index='config', column='replica')

        # Return Hansmann parameter
        return 1. - np.sqrt(np.square(data).sum(axis=1)) / data.sum(axis=1)

    def hansmann_plot(self, x_title=None, y_title=None, height=None, width=None, plot_theoretical=True):
        """
        Plot the Hansmann parameter.

        Parameters
        ----------
        x_title : str
        y_title : str
        height : numeric
            Height of plot.
        width : numeric
            Width of plot.
        plot_theoretical : bool
            Should the theoretical Hansmann parameter in the case of equal sampling be plotted? (Default: True)
        """

        import uplot as u

        data = self.hansmann()
        x = data.index.to_numpy()
        if x[0] == 0:  # start everything from an index of 1 for aesthetics
            x += 1
        y = data.to_numpy()

        # Build figure
        fig = u.figure(style={
            'x_title': x_title if x_title else r'replica index, $r$',
            'y_title': y_title if y_title else r'mixing parameter, $m$($r$)',
            'y_min': 0.,
            'y_max': 1.,
            'height': height,
            'width': width
        })
        fig += u.line(x, y, style={'color': 'black', 'line_style': 'solid', 'marker': 'circle'})  # noqa
        if plot_theoretical:
            fig += u.hline(
                y=1. - 1. / np.sqrt(len(x)),  # noqa
                style={'color': 'black', 'line_style': 'longdash', 'line_width': 1.}
            )
        fig, ax = fig.to_mpl(show=False)
        fig.savefig('hansmann_plot.svg')

    def mosaic_plot(self, interval=100, cmap='ujet', height=None, width=None):
        """
        Plot the mosaic function using a heatmap.

        Parameters
        ----------
        interval : int
            Interval for plotting. We usually cannot plot every single step because there is too much data.
        cmap : str or object
            The matplotlib-compatible color map.
        height : numeric
            Height of plot.
        width : numeric
            Width of plot.
        """

        import matplotlib.pyplot as plt
        from matplotlib.ticker import MultipleLocator, MaxNLocator
        import uplot as u

        u.core.set_mpl_theme()

        data = self.trajectory(by='config', reset_index=True)
        steps = data.index.to_numpy(dtype='int')[::interval]
        replicas = data.columns.to_numpy(dtype='int')
        if replicas[0] == 0:
            replicas = replicas + 1
            data = data + 1
        mosaic = data.to_numpy(dtype='int')[::interval, :]
        x = np.arange(mosaic.shape[0] + 1)
        y = np.arange(mosaic.shape[1] + 1)

        # Start figure and axis
        figsize = None
        if height is not None and width is not None:
            figsize = (width, height)
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot()
        # im = ax.pcolormesh(mosaic, cmap=cmap, edgecolors='k', linewidth=0.5)  # bwr
        if cmap == 'ujet':
            import uplot
            cmap = uplot.jet
        im = ax.pcolormesh(x - 0.5, y - 0.5, mosaic.T, cmap=cmap, edgecolors='k', linewidth=0.5)

        # Format x axis
        # TODO change this so only units of X are display
        ax.set_xticks(np.arange(len(steps)))
        ax.set_xticklabels(steps)
        ax.set_xlabel(r'step')

        # Format y axis
        ax.set_yticks(np.arange(len(replicas)))
        ax.set_yticklabels(replicas)
        ax.set_ylabel(r'temperature index')

        # Format tick lines
        # ax.spines['top'].set_visible(False)
        # ax.spines['right'].set_visible(False)
        ax.xaxis.set_ticks_position('bottom')
        ax.yaxis.set_ticks_position('left')
        ax.tick_params(axis='both', which='both', direction='out')

        ax.xaxis.set_major_locator(MultipleLocator(10))
        ax.xaxis.set_minor_locator(MultipleLocator(10))

        ax.yaxis.set_major_locator(MultipleLocator(1))
        ax.yaxis.set_minor_locator(MultipleLocator(1))

        # fig.colorbar()
        ax.set_aspect('auto')
        # ax.grid(which='minor', color='w', linestyle='-', linewidth=5)

        ax.grid(linestyle='')

        # plt.axis('equal')
        # Add color bar
        cbar = plt.colorbar(im, ax=ax, shrink=0.5, drawedges=False)
        cbar.outline.set_linewidth(0.5)
        # cbar.ax.spines['right'].set_visible(True)
        cbar.ax.tick_params(direction='out', length=5.)
        cbar.ax.tick_params(which='minor', length=0)
        #        cbar.ax.yaxis.set_major_locator(MultipleLocator(1))
        cbar.ax.yaxis.set_major_locator(MaxNLocator(integer=True))
        cbar.ax.set_ylabel(r'replica index')

        # Save the image
        # fig.show()
        fig.savefig('mosaic_plot.svg')

    # Exchange rate plot
    def rate_plot(self):
        import uplot as u
        u.plot(
            self.exchange_rate(by='config'),  # noqa
            x_title='replica',
            y_title='exchange rate',
            y_min=0.,
            y_max=1.,
            height=5,
            width=6,
            save_as='exchange_rate.svg',
        )

    # Reset the step index?
    def reset_step(self):
        """
        Resets step from arbitrary interval to rank from 1 ... N. This operation occurs in place.
        """

        self._data['step'] = self._data['step'].rank(method='dense')

    # Save to csv
    def to_csv(self, *args, **kwargs):
        """
        Save ExchangeHistory to csv format. Follows :ref:`pandas.DataFrame.to_csv`.
        """
        self._data.to_csv(*args, **kwargs)

    # Save to parquet
    def to_parquet(self, *args, **kwargs):
        """
        Save ExchangeHistory to parquet format. Follows :ref:`pandas.DataFrame.to_parquet`.
        """

        self._data.to_parquet(*args, **kwargs)

    # TODO should this be renamed mosaic?
    def trajectory(self, by='config', reset_index=True):
        """
        Compute the walk of config or replicas. This is a pivot of the melted ExchangeHistory._data.

        Parameters
        ----------
        by : str
            Default: config.
        reset_index : bool
            Should the index be reset?

        Returns
        -------
        pandas.DataFrame
        """

        # Define the columns and values for the pivot
        columns = 'config'
        values = 'replica'
        if by == 'replica':
            columns, values = values, columns
        elif by != 'config':
            raise AttributeError(f'do not understand by = {by}')

        # Perform the pivot
        data = self._data.pivot(index='step', columns=columns, values=values)

        # Reset index?
        if reset_index:
            data.reset_index(drop=True, inplace=True)
            data.index.name = 'step'

        # Return
        return data

    # State transitions
    def state_transitions(self, walker='replica'):
        """
        Compute state transitions similar to the "tunneling time" as referenced in [Straub09JCP]_. The language here is
        somewhat different than the rest of the class. We have a "walker" that randomly walks across "state" space.
        We are counting transitions of length :math:`r`, which occurs when a walker originally at state :math:`i`
        moves to the state :math:`i+r` or :math:`i-r`.

        Consider a walker samples 3 states given by indices 0-2.

        .. code-block::
           00011000112210012

        This can be decomposed in several ways. For instance, if :math:`i=0` and :math:`r=1`, then there are 3 such
        transition events: "0001-------------", "-----0001--------", and "-------------001-". For clarity,
        the states not involved explicitly in this type of transition are denoted by "-". Notice that we are always
        looking at the first instance of state :math:`i` and the steps after :math:`i` it takes to get to the first
        instance of state :math:`i+r` (or :math:`i-r`). We can do this again where :math:`i=1`, :math:`i+r=2`,
        and :math:`i-r=0`. The transition events are "---110-----------", "--------112------", "------------10---",
        and "---------------12". Taken together, for transitions of length :math:`r=1`, we have 7 events: 2 with 4
        steps, 3 with 3 steps, and 2 with 2 steps. We can therefore say that the  average number of steps
        for that transition length is :math:`3`.

        Returns
        -------
        dict of arrays
            The key in the dictionary is all possible transition lengths :math:`1...R-1`, where :math:`R` is the
            number of replicas.

        References
        ----------
        .. [Straub09JCP] Kim, J., Keyes, T., & Straub, J. E. (2009) Replica exchange statistical temperature Monte
        Carlo. *J. Chem. Phys.* 130, 124112.
        """

        data = self.trajectory(walker)
        if not (data.diff().abs().max() == 1.).all():  # noqa
            raise AttributeError(f'{walker} cannot be the walker, step size exceeds 1')

        # TODO can this be simplified?
        n_steps, n_walkers = data.shape
        n_states = n_walkers  # redundant, but makes logic clearer
        steps = {}
        for k in range(n_walkers):  # This is the walker
            walker_data = data[walker].to_frame('state')
            for state in range(n_states):  # First state
                for r in range(1, n_states - 1):  # To get either state - r or state + r
                    state_data = walker_data.query(f'state in [{state}, {state - r}, {state + r}]').reset_index()
                    mask = state_data['state'].diff() != 0  # remove repeated rows
                    state_data = state_data[mask].copy()  # apply mask
                    state_data['n_steps'] = state_data['step'].shift(-1) - state_data['step']
                    transitions = state_data.query(f'state == {state}')['n_steps'].dropna().to_numpy()  #
                    steps[r] = np.append(steps.get(r, []), transitions)

        return steps


# Replica state
class State:
    def __init__(self):
        self.temperature = None
        self.pressure = None


# Create a temperature schedule
# Variable names taken from numpy.linspace
def temperature_schedule(start=300, stop=440, num=40, mode='geometric'):
    r"""
    Create a temperature schedule that could be used, for instance, with replica exchange.

    There are several choices for `mode`. Note that :math:`T` refers to the temperature at :math:`i = 1 ... R`, where
    :math:`T_1` is `start` and :math:`T_R` is `stop`. In total, there are :math:`R` temperatures (= `num`).

    * "geometric" [#]_

    .. math :: T_i = T_1 \left( \frac{T_R}{T_1} \right)^{\frac{i-1}{R-1}}

    * "linear"

    .. math :: T_i = T_1 + (i-1) \frac{T_R-T_1}{R-1}

    * "parabolic" (Note if `num` is even, `stop` won't directly be sampled).

    .. math :: T_i = T_1 - \frac{T_R-T_1}{\left( \frac{R-1}{2} \right) ^2} (i-1) (i-R)

    Parameters
    ----------
    start : float
        Lowest temperature
    stop : float
        Highest temperature
    num : int
        Number of temperatures
    mode : str
        Mode to produce schedule. Valid options include "geometric", "linear", "parabolic". Any substring will match,
        but the preference should be to use the full option label. (Default: "geometric")

    Returns
    -------
    numpy.ndarray
        Temperature schedule

    Examples
    --------
    .. plot::
       :include-source:

       import matplotlib.pyplot as plt
       import molecular as mol

       num = 10
       geometric = mol.temp_schedule(300, 440, num, 'geometric')
       linear = mol.temp_schedule(300, 440, num, 'linear')
       parabolic = mol.temp_schedule(300, 440, num, 'parabolic')

       plt.figure()
       plt.plot(range(num), geometric, label='geometric')
       plt.plot(range(num), linear, label='linear')
       plt.plot(range(num), parabolic, label='parabolic')
       plt.xlabel('index')
       plt.ylabel('temperature')
       plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
       plt.show()

    References
    ----------
    .. [#] Nymeyer, H., Gnanakaran, S., & García, A. E. (2004) Atomistic simulations of protein folding, using the
       replica exchange algorithm. *Methods Enzymol.* **383**: 119-149.
    """

    mode = mode.lower()

    if mode in 'geometric':
        schedule = start * np.power(stop / start, np.arange(num) / (num - 1.), dtype=np.float64)

    elif mode in 'linear':
        schedule = np.linspace(start=start, stop=stop, num=num)

    elif mode in 'parabolic':
        temp_range = stop - start
        temp_ind = np.arange(num)
        schedule = start - (temp_range / np.square((num - 1.) / 2.)) * temp_ind * (temp_ind - num + 1.)

    else:
        raise AttributeError(f'mode {mode} not supported')

    return schedule


# Convenience rename
tempsch = temperature_schedule


if __name__ == '__main__':
    print(temperature_schedule(start=310, stop=510, num=8, mode='geometric'))
    # print(temp_schedule(start=310, stop=500, num=5, mode='parabolic'))



"""
energy.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

from molecular.geometry import *

from abc import ABCMeta
import numpy as np


# TODO allow computation to work with only distances? Need a way to speed up calculations using REMD

def lennard_jones(epsilon, r_min, r):
    return epsilon * ((r_min / r) ** 12. - 2 * (r_min / r) ** 6.)


# Metaclass that defines structure of each energy term
class EnergyTerm(metaclass=ABCMeta):
    """
    Metaclass that defines structure of each energy term.
    """

    # Initialize class instance
    def __init__(self):
        self._energy = None
        self._force = None

    def _compute_energy(self):
        raise NotImplementedError

    def _compute_force(self):
        raise NotImplementedError

    @property
    def energy(self):
        """
        The computed energy term

        Returns
        -------
        numpy.ndarray
            Energy
        """

        if self._energy is None:
            self._energy = result = self._compute_energy()
        else:
            result = self._energy
        return result

    @property
    def force(self):
        """
        The computed force vector

        Returns
        -------
        numpy.ndarray
            Force vector
        """

        if self._force is None:
            result = self._force = self._compute_force()
        else:
            result = self._force
        return result


# Bond energy term
class Bond(EnergyTerm):
    r"""
    Compute the bond potential energy according to Hooke's law

    The energy :math:`U` can be computed from the difference in the measured distance :math:`r` from the ideal distance
    :math:`r_0`, scaled by the spring constant :math:`k`.

    .. math:: U = \frac{1}{2} k (r - r_0)^2

    Since :math:`r` can be separated out into components :math:`x`, :math:`y`, and :math:`z`, we must compute the
    partial derivatives of :math:`\delta U`. Then, :math:`F = -\delta U / \delta \Delta x`.

    .. math:: F = -\frac{\delta U}{\delta \Delta x} = -k(r-r_0)\frac{\Delta x}{r}

    For the derivation:

    .. math:: \delta U = k(r-r_0) \delta r
    .. math:: r = \sqrt{r^2}
    .. math:: \delta r = \frac{1}{2}(r^2)^{-1/2} \delta r^2
    .. math:: r^2 = (\Delta x)^2 + (\Delta y)^2 + (\Delta z)^2
    .. math:: \frac{\delta r^2}{\delta \Delta x} = 2 \Delta x

    The same for above can be computed for :math:`y` or :math:`z`.

    Examples
    --------
    >>> from molecular import Bond
    >>> bond = Bond(atom1, atom2, idval)
    >>> bond.energy
    >>> bond.derivative
    >>> bond.force
    """

    def __init__(self, a, b, ideal_value=1., force_constant=1.):
        """
        Initialize instance of Bond class.

        Parameters
        ----------
        a, b : ArrayLike
            Cartesian coordinates
        ideal_value : ArrayLike
            The ideal distance between the particles.
        force_constant : ArrayLike
            The spring constant describing the particles.
        """

        # Coerce shapes
        a, b = _coerce_xyz(a, b)

        # Compute
        self._vector = vector(a, b)
        self._distance = distance(a, b)
        self._offset = self._distance - ideal_value

        # Save parameters
        EnergyTerm.__init__(self)
        self._ideal_value = ideal_value
        self._force_constant = force_constant

    def _compute_energy(self):
        return 0.5 * self._force_constant * np.square(self._offset)

    def _compute_force(self):
        return -1. * self._force_constant * self._offset[:, None] * self._vector / self._distance[:, None]


class Angle(EnergyTerm):
    r"""
    Compute the potential energy of an angle.

    The energy of an angle follows the harmonic potential:

    .. math:: U = \frac{1}{2} k (\theta - \theta_0)^2

    The derivative must be split by the components of vectors :math:`u` and :math:`v`, which are used to compute the
    angle :math:`\theta`:

    .. math:: F = -\frac{\delta U}{\delta x} = k (\theta - \theta_0) sin^{-1}(\theta) / |v|



    .. math:: \frac{\delta U}{\delta v_x} = -k (\theta - \theta_0) sin^{-1}(\theta) / |u|

    The derivation:

    .. math:: \frac{\delta U}{\delta \theta} = k (\theta - \theta_0)
    .. math:: \theta = acos(\hat{u_{ij}} \cdot \hat{u_{jk}})
    .. math:: \frac{\delta \theta}{\delta u \cdot v} = \frac{-1}{\sqrt{1 - (\hat{u} \cdot \hat{v})^2}}
                                               = \frac{-1}{\sqrt{1 - cos^2(\theta)}}
                                               = \frac{-1}{\sqrt{sin^2(\theta)}}
                                               = \frac{-1}{sin(\theta)}
                                               = sin^{-1}(\theta)

    .. math:: u_i = [x, y, z]
    .. math:: \frac{\delta u_i}{\delta x} = 1

    .. math:: u \cdot v = u_x v_x + u_y v_y + u_z v_z
    .. math:: \delta (\^{u} \cdot \^{v}) = u \frac{dv}{dx} v \frac{du}{dx}

    .. math:: \delta (\^{u} \cdot \^{v}) = u \frac{dv}{dx} v \frac{du}{dx}
    .. math:: u = [\Delta x, \Delta y, \Delta z]

    f1 = dth/duv / |u| * (u * (cos th / |u|) - v / |v|) = dth/duv / |u| * (u * costh - v)
    f3 = dth/duv / |v| * (v * (cos th / |v|) - u / |u|) = dth/duv / |v| * (v * costh - u)
    f2 = -1 * (f1 + f3)


    .. math:: \frac{\delta u}{\delta \Delta x} = [1, 1, 1]

    .. math:: \frac{\delta (u \cdot v)}{\delta u_x} = v_x
    .. math:: \frac{\delta (u \cdot v)}{\delta v_x} = u_x

    Parameters
    ----------
    xyz0
    xyz1
    xyz2
    ideal_value
    force_constant

    Returns
    -------

    """

    def __init__(self, xyz0, xyz1, xyz2, ideal_value=1., force_constant=1.):
        """
        Parameters
        ----------
        xyz0 : ArrayLike
            Cartesian coordinates for first particle.
        xyz1 : ArrayLike
            Cartesian coordinates for second particle.
        xyz2 : ArrayLike
            Cartesian coordinates for third particle.
        ideal_value : ArrayLike
            The ideal angle between the particles.
        force_constant : ArrayLike
            The spring constant describing the particles.
        """

        # Coerce shapes
        xyz0, xyz1, xyz2 = _coerce_xyz(xyz0, xyz1, xyz2)

        # Compute angle
        self._angle = angle(xyz0, xyz1, xyz2)
        self._offset = self._angle - ideal_value

        # Save parameters
        EnergyTerm.__init__(self)
        self._ideal_value = ideal_value
        self._force_constant = force_constant

    def _compute_energy(self):
        return 0.5 * self._force_constant * np.square(self._offset)

    def _compute_force(self):
        pass


class Dihedral(EnergyTerm):
    pass


# TODO LJ with pairlists
class LennardJones(EnergyTerm):
    r"""

    .. math:: U = 4\epsilon [(\frac{\sigma}{r})^{12} - (\frac{\sigma}{r})^6]
    .. math:: F = -\frac{\delta U}{\delta x} = -24\frac{\epsilon \Delta x}{r}[2(\frac{\sigma}{r})^{12} - (\frac{\sigma}{r})^6]

    .. math:: U = 4\epsilon [z^{12} - z^6]
    .. math:: dU/dx = 4e [12z^{11} - 6z^5] dz
    .. math:: dU/dx = 24e [2z^{11} - z^5] dz
    .. math:: z = s / r = sr^{-1}
    .. math:: dz/dx = -sr^{-2} dr
    .. math:: r = sqrt(delta x + delta y + delta z)
    .. math:: dr/dx = 1/2(r^2)^(-1/2) * d(r^2)
    .. math:: r2 = delta x + delta y + delta z
    .. math:: d(r^2)/dx = delta x


    .. math:: \frac{\delta U}{\delta x} = 4\epsilon [12(\frac{\sigma}{r})^{11} - 6(\frac{\sigma}{r})^5] *
    """

    def __init__(self, xyz0, xyz1, sigma=1., epsilon=-1):
        """
        Parameters
        ----------
        xyz0 : ArrayLike
            Cartesian coordinates for first particle.
        xyz1 : ArrayLike
            Cartesian coordinates for second particle.
        sigma : float
        epsilon : float
        """

        # Coerce shapes
        xyz0, xyz1 = _coerce_xyz(xyz0, xyz1)

        # Compute distance
        self._vector = vector(xyz0, xyz1)
        self._distance = norm(self._vector)
        if self._distance < 1e-7:
            r = 1000.
        else:
            r = sigma / self._distance
        self._r6 = np.power(r, 6)
        self._r12 = np.power(self._r6, 2)

        # Save parameters
        EnergyTerm.__init__(self)
        self._sigma = sigma
        self._epsilon = epsilon

    # TODO save some of these intermediary calculations for force
    def _compute_energy(self):
        return 4. * self._epsilon * (self._r12 - self._r6)

    def _compute_force(self):
        return -24. * self._vector * self._epsilon / self._distance * (2. * self._r12 - self._r6)


class FullElectrostatic(EnergyTerm):
    pass



# Coerce coordinates into an acceptable form
def _coerce_xyz(*args):
    # Place to store results
    results = []

    # Loop over all coordinates in arguments
    shape = None
    for i in range(len(args)):
        # Make sure we're dealing with a 2D array
        xyz = np.array(args[i])
        if xyz.ndim == 1:
            xyz = xyz.reshape(1, -1)

        # Make sure coordinates are the same shape
        if shape is None:
            shape = xyz.shape
        elif shape != xyz.shape:
            raise AttributeError('coordinates must be same shapes')

        # Save coerced coordinates
        results.append(xyz)

    # Return
    return results


class HarmonicOscillator:
    """

    """

    __slots__ = ('_ideal_value', '_force_constant')

    def __init__(self, ideal_value, force_constant):
        self._ideal_value = ideal_value
        self._force_constant = force_constant

    def energy(self, instantaneous_value):
        return 0.5 * self._force_constant * np.square(instantaneous_value - self._ideal_value)

    def force(self, instantaneous_value):
        return self._force_constant * (instantaneous_value - self._ideal_value)


"""
statistics.py

author: C. Lockhart <clockha2@gmu.edu>
language: Python3
"""

from scipy.stats import sem as _sem


# Compute SEM
def sem(a, *args, **kwargs):
    r"""
    Compute standard error of mean, :math:`\delta_M`. See :ref:`scipy.stats.sem`.

    .. math::

       \delta_{M} = \frac{\sqrt{\frac{1}{n-1} \sum_{i=1}^n (x(i) - \mu_x)^2}}{\sqrt{n}}

    This is equivalent to np.std(a, ddof=1) / np.sqrt(len(a))
    """

    return _sem(a, *args, **kwargs)





#cython: language_level=3

from cython.parallel cimport prange
from scipy.spatial.distance import cdist
import numpy as np
cimport numpy as np


def _distances(a, b, parallelize=False):
    r = np.zeros((a.shape[0], a.shape[1], b.shape[1]))
    if not parallelize:
        for i in range(a.shape[0]):
            r[i, :, :] = cdist(a[i, :, :], b[i, :, :])  # sklearn also has pairwise_distances
    else:
        raise AttributeError  # not implemented for now
        # for i in prange(a.shape[0], nogil=True):
        #     r[i, :, :] = cdist(a[i, :, :], b[i, :, :])
    return r


def _minimum_distances(a, b, box):
    # Convert box to right format
    box = box[:, np.newaxis, :]

    # Put a and b in box
    a = a - box * np.round(a / box)
    b = b - box * np.round(b / box)

    # Find minimum distances
    min_r = np.ones((a.shape[0], a.shape[1], b.shape[1])) * np.inf
    for dx in range(-1, 2):
        for dy in range(-1, 2):
            for dz in range(-1, 2):
                # Move b
                bm = b + box * [dx, dy, dz]

                # Update min_r?
                r = _distances(a, bm)
                is_closer = r < min_r
                if np.sum(is_closer) > 0:
                    min_r[is_closer] = r[is_closer]

    # Return
    return min_r




"""
distance.py

language: python
version: 3.x
author: C. Lockhart <chris@lockhartlab.org>
"""

from molecular.analysis._analysis_utils import _distances, _minimum_distances

from iox import collapse_to_vector, numpy_aggregate
import numpy as np
import pandas as pd
from sparse import COO  # has numba dependency I'm not thrilled about


def contacts(a, b, cutoff=4.5, aggregate_by=None, include_images=False, vectorize=False):
    """
    Compute atomic contacts.

    Parameters
    ----------
    a, b : Trajectory
    cutoff : float
    aggregate_by : str
        Aggregate atomic contacts to a different resolution, e.g., residue or segment.
    include_images : bool
        Include neighboring images in this calculation (Default: False).
    vectorize : bool

    Returns
    -------
    numpy.ndarray
    """

    # Actually compute the contacts
    results = distances(a, b, include_images=include_images) <= cutoff

    # Aggregate the contacts?
    if aggregate_by is not None:
        results = _format_results(results, a, b, value_name='contact', by=aggregate_by, aggfunc='max')

    # Vectorize?
    if vectorize:
        results = collapse_to_vector(results, index='structure_id', where='contact == True').drop(columns='contact')

    # Return
    return results


def contacts_to_vector(contacts, axis1=None, axis2=None):
    """
    Convert contacts array to contact vector. With `axis1` and `axis2`, `contacts` can be changed such that indices
    besides the atom IDs are used.

    Parameters
    ----------
    contacts : numpy.ndarray
    axis1 : list-like
        (Optional) New indices for axis 1 of `contacts`
    axis2 : list-like
        (Optional) New indices for axis 2 of `contacts`

    Returns
    -------
    numpy.ndarray
    """

    # Get sparse representation of contacts
    contacts_sparse = COO(contacts)

    # Update axis IDs if necessary
    def _update_id(i, x):
        contacts_sparse.coords[i] = \
            pd.Series(dict(zip(range(contacts.shape[i]), x)))[contacts_sparse.coords[i]].to_numpy()

    if axis1 is not None:
        _update_id(1, axis1)
    if axis2 is not None:
        _update_id(2, axis2)

    # Create DataFrame
    df = pd.DataFrame({
        'structure_id': contacts_sparse.coords[0],
        f'i': contacts_sparse.coords[1],
        f'j': contacts_sparse.coords[2],
    })

    # Aggregate to series
    sr = (
        df
        .drop_duplicates()
        .sort_values(['structure_id', 'j'])
        .groupby('structure_id')['j']
        .agg(list)
    )

    # Make a correction for missing structure IDs (if there were no contacts they'll be absent)
    is_empty = ~np.max(contacts, axis=(1, 2))
    if np.sum(is_empty) > 0:
        for i in np.argwhere(is_empty).ravel():
            sr.loc[i] = []

        sr.sort_index(inplace=True)

    # Return as numpy array
    return sr.to_numpy()


# Compute the distance between two Trajectories
def distances(a, b, include_images=False):
    """
    Compute the distance between two Trajectory instances.

    Parameters
    ----------
    a, b : Trajectory
        Two trajectories. Must have same dimensions.
    include_images : bool
        Consider all images in this computation. (Default: False)

    Returns
    -------
    numpy.ndarray
        Distance between every frame in the trajectory.
    """

    # Extract coordinates
    a_xyz = a.xyz.to_numpy().reshape(*a.shape)
    b_xyz = b.xyz.to_numpy().reshape(*b.shape)

    # If we're not going to consider images, we can run with just coordinates
    if not include_images:
        return _distances(a_xyz, b_xyz)

    # Otherwise, we need to do some extra checks and extract the box
    else:
        # Make sure that `a` and `b` come from the same parent Trajectory
        if a.parent.designator != b.parent.designator:
            raise AttributeError('`a` and `b` must have same parent')

        # Extract boxes and check that `a_box` and `b_box` are identical
        a_box = a.box.to_numpy()
        b_box = b.box.to_numpy()
        if np.allclose(a_box, b_box) is not True:
            raise AttributeError('`a` and `b` have different dimensions')

        # Finally, we can compute the minimum distances
        return _minimum_distances(a_xyz, b_xyz, a_box)


#
# # Compute the distance between two Trajectories
# def distance(a, b):
#     """
#     Compute the distance between two Trajectory instances.
#
#     Parameters
#     ----------
#     a, b : Trajectory
#         Two trajectories. Must have same dimensions.
#
#     Returns
#     -------
#     numpy.ndarray
#         Distance between every frame in the trajectory.
#     """
#
#     # TODO there must be a better way
#     a_xyz = a.xyz.to_numpy().reshape(*a.shape)
#     b_xyz = b.xyz.to_numpy().reshape(*b.shape)
#
#     return np.sqrt(np.sum(np.square(a_xyz - b_xyz), axis=(1, 2)))
#
# # Compute pairwise distance between two Trajectories (or within a Trajectory?)
# def pairwise_distance(a, b):
#     pass



def _format_results(results, a, b, value_name, aggfunc='min', by='residue'):
    axes = (
        None,
        a[f'{by}_id'],
        b[f'{by}_id']
    )

    index_names = (
        'structure_id',
        f'{by}_id0',
        f'{by}_id1'
    )

    return numpy_aggregate(
        results,
        axes=axes,
        index_names=index_names,
        value_name=value_name,
        aggfunc=aggfunc,
        as_frame=True
    )





from molecular.analysis import contacts
from molecular.simulations import generate_images

import numpy as np


def has_cross_interactions(a, cutoff=4.5):
    # Move `a` to the unit cell
    am = a.to_origin(inplace=False)

    # Go through all images and find cross interactions
    is_crossed = np.zeros(a.n_structures, dtype='bool')
    for image in generate_images():
        if image == (0, 0, 0):  # skip the origin
            continue
        bm = a.to_image(*image, inplace=False)  # use `a` directly to avoid creating a copy
        is_crossed = is_crossed | np.max(contacts(am, bm, cutoff=cutoff, include_images=False), axis=(1, 2))

    # Return
    return is_crossed







import numpy as np
from numpy.distutils.core import setup
from numpy.distutils.misc_util import Configuration


# Create configuration
def configuration(parent_package='', top_path=None):
    config = Configuration('analysis', parent_package, top_path)
    config.add_extension(
        '_analysis_utils',
        sources=['_analysis_utils.pyx'],
        include_dirs=[np.get_include()]
    )
    return config


setup(configuration=configuration)


"""
trajectory.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

from molecular.core.errata import pivot
from molecular.transform import center, move

import logging
import numpy as np
import pandas as pd
from typelike import ArrayLike, NumberLike
from tqdm import tqdm

# Get the molecular.core logger
logger = logging.getLogger('molecular.core')


# https://softwareengineering.stackexchange.com/questions/386755/sharing-docstrings-between-similar-functions
def set_doc(original):
    def wrapper(target):
        target.__doc__ = original.__doc__.replace('a : molecular.Trajectory', '')
        return target

    return wrapper


# Trajectory class
class Trajectory(object):
    """
    `Trajectory` stores instantaneous data for a `Topology`, i.e., atomic coordinates, box information, velocity, etc.


    data:
        - structure_id
        - atom_id
        - x
        - y
        - z
        - vx
        - vy
        - vz
        - fx
        - fy
        - fz
        - alpha
        - beta
        - user

    config
        - structure_id
        - bx
        - by
        - bz


    """

    __slots__ = ['_data', '_configuration', '_topology', '_parent']

    @classmethod
    def from_universe(cls, universe):
        return from_universe(universe, return_topology=False, return_trajectory=True)

    # Initialize instance of Trajectory
    def __init__(self, data=None, coordinates=None, box=None, topology=None):
        """


        Parameters
        ----------
        data : pandas.DataFrame
            (Optional) Structured information about Trajectory. Columns must include "structure_id", "atom_id", "x",
            "y", and "z".
        coordinates
        box
        topology : Topology
            (Optional) Topology that includes information about atoms.
        """

        # Make sure data has correct columns if available
        if isinstance(data, pd.DataFrame):
            data.reset_index(inplace=True)
            if not np.in1d(['structure_id', 'atom_id', 'x', 'y', 'z'], data.columns).any():
                raise AttributeError('data must include structure_id, atom_id, x, y, and z')
            data.set_index(['structure_id', 'atom_id'], inplace=True)
        elif data is not None:
            raise AttributeError('data must be pandas.DataFrame')

        # Process coordinates if set
        if coordinates is not None and isinstance(coordinates, np.ndarray):
            if data is None:
                data = _dummy_trajectory_data(*coordinates.shape)
            data[['x', 'y', 'z']] = coordinates.reshape(-1, 3)

        # Process box if set
        # TODO other names could be metadata... Ensemble class? System class? these names are bad SimulationParameters?
        configuration = None
        if box is not None and isinstance(box, np.ndarray):
            configuration = pd.DataFrame({
                'structure_id': np.arange(box.shape[0]),
                'bx': box[:, 0],
                'by': box[:, 1],
                'bz': box[:, 2]
            })
            configuration.set_index('structure_id', inplace=True)

        # Process topology
        if topology is not None:
            if isinstance(topology, Trajectory):
                topology = topology.topology
            elif not isinstance(topology, Topology):
                raise AttributeError('topology must be Topology instance')

        # Save Trajectory data elements
        self._data = data
        self._configuration = configuration
        self._topology = topology
        self._parent = None

    # Add
    def __add__(self, other):
        if isinstance(other, NumberLike):
            self._data[['x', 'y', 'z']] = self._data[['x', 'y', 'z']] + other
        return self

    # Get item
    def __getitem__(self, item):
        """


        Parameters
        ----------
        item : str

        Returns
        -------

        """

        # NOTE this used to slide _data index... might be some bugs down the road

        # Look in Trajectory first
        if item in self._data.columns:
            return self._data[item]

        # Then look in Topology
        elif item in self._topology._data.columns:
            return self._topology._data[item]

        # If we still can't find the item, we've failed
        else:
            raise AttributeError(f'{item} not found')

    # Length of trajectory
    def __len__(self):
        """
        The length of the trajectory is the number of structures.

        Returns
        -------
        int
            Length of trajectory
        """

        return self.n_structures

    def __mul__(self, other):
        pass

    # Representation of the object
    def __repr__(self):
        return "# structures: {0}\n# atoms: {1}\n# dimensions: {2}".format(*self.shape)

    # Subtract the coordinates of this Trajectory by another coordinate system
    def __sub__(self, other):
        if isinstance(other, Trajectory):
            other = other.coordinates
        return self.move(by=-1. * other, inplace=False)

    # Get unique atom IDs (can be cached)
    @property
    def atom_ids(self) -> np.ndarray:
        """
        Get unique atom IDs in the Trajectory.

        Returns
        -------
        numpy.ndarray
            Array of atom IDs.
        """

        # noinspection PyUnresolvedReferences
        return self._data.index.levels[1]

    # Box
    @property
    def box(self):
        return self._configuration  #.set_index('structure_id')[['bx', 'by', 'bz']]

    # Get coordinates
    @property
    def coord(self):
        return self.coordinates

    # Set coordinates
    @coord.setter
    def coord(self, coord):
        self.coordinates = coord

    # Get coordinates
    @property
    def coordinates(self):
        """
        Get Cartesian coordinates from the Trajectory.

        Returns
        -------
        pandas.DataFrame
            Trajectory Cartesian coordinates.
        """

        return self._data[['x', 'y', 'z']]

    # Set coordinates
    @coordinates.setter
    def coordinates(self, coordinates):
        """
        Set Cartesian coordinates for the Trajectory.

        Parameters
        ----------
        coordinates : array-like
            Cartesian coordinates of same shape as Trajectory.
        """

        # Set coordinates
        self._data[['x', 'y', 'z']] = coordinates

    @property
    def columns(self):
        """

        Returns
        -------
        numpy.ndarray
        """

        return self._data.columns.to_numpy()

    # Trajectory designator
    @property
    def designator(self):
        return f'Trajectory:{self.hex_id}'

    # Hex ID of Trajectory
    @property
    def hex_id(self):
        return hex(id(self))

    # Number of atoms
    @property
    def n_atoms(self):
        """
        Number of atoms in the `Trajectory`.

        Returns
        -------
        int
            Number of atoms
        """

        # noinspection PyUnresolvedReferences
        return self._data.index.levshape[1]

    # Number of dimensions
    @property
    def n_dim(self):
        """
        Number of dimensions.

        Returns
        -------
        int
            Number of dimensions
        """

        n_dim = 0
        if 'x' in self._data.columns: n_dim = n_dim + 1
        if 'y' in self._data.columns: n_dim = n_dim + 1
        if 'z' in self._data.columns: n_dim = n_dim + 1

        return n_dim

    @property
    def n_residues(self):
        return self._topology.n_residues

    # Number of structures
    @property
    def n_structures(self):
        """
        Number of structures in the `Trajectory`.

        Returns
        -------

        int
            Number of structures
        """

        # noinspection PyUnresolvedReferences
        return self._data.index.levshape[0]

    # Parent
    @property
    def parent(self):
        """
        Parent of the Trajectory.

        Returns
        -------
        Trajectory or None
        """

        return self._parent

    # Shape
    @property
    def shape(self):
        """
        Shape of `Trajectory`

        Returns
        -------
        tuple
            (Number of structures, number of atoms, dimensionality)
        """

        return self.n_structures, self.n_atoms, self.n_dim

    # Get a list of structures
    @property
    def structure_ids(self) -> np.ndarray:
        """
        Get a numpy array of structure indices.

        Returns
        -------
        numpy.ndarray
            Structure indices.
        """

        # noinspection PyUnresolvedReferences
        return self._data.index.levels[0]

    # Get topology
    @property
    def topology(self):
        """

        Get the `Topology` instance.

        Returns
        -------
        Topology
            `Topology` instance associated with this `Trajectory`.
        """

        # Check the topology
        self._check_topology()

        # Return
        return self._topology

    # Get x
    @property
    def x(self):
        """
        Get x coordinates.

        Returns
        -------
        numpy.ndarray
            x coordinates.
        """

        return self._data['x']

    # Get y
    @property
    def y(self):
        """
        Get y coordinates.

        Returns
        -------
        numpy.ndarray
            y coordinates.
        """

        return self._data['y']

    # Get z
    @property
    def z(self):
        """
        Get z coordinates.

        Returns
        -------
        numpy.ndarray
            z coordinates.
        """

        return self._data['z']

    # Get xyz coordinates
    @property
    def xyz(self):
        """
        Get x, y, and z coordinates.

        Returns
        -------
        pandas.DataFrame
            Cartesian coordinates.
        """

        return self.coordinates

    @xyz.setter
    def xyz(self, xyz):
        self.coordinates = xyz

    # Check topology
    def _check_topology(self):
        """
        Check that the topology is correctly set
        """

        # First off, must be Topology instance
        if not isinstance(self._topology, Topology):
            raise AttributeError('topology is not correct class (%s)' % type(self._topology))

        # Second, number of atoms must match
        if self.n_atoms != self._topology.n_atoms:
            raise AttributeError('number of atoms in topology and trajectory do not match ({0} vs {1})'
                                 .format(self.n_atoms, self._topology.n_atoms))

    # Apply
    # TODO please optimize this. Use vectorize, map, parallelization
    def apply(self, function, progress_bar=False):
        result = []
        structure_ids = np.arange(self.n_structures)
        if progress_bar:
            structure_ids = tqdm(structure_ids)
        for i in structure_ids:
            structure = self.get_structure(i)
            result.append(function(structure))
        return result

    # Compute the center of the Trajectory
    @set_doc(center)
    def center(self, weights=None):
        return center(self, weights=weights)

    # Copy
    # TODO
    def copy(self):
        """
        Create a copy of the Trajectory.

        Returns
        -------
        molecular.Trajectory
            Deep copy of the Trajectory.
        """

        return Trajectory(
            data=self._data.copy(),
            box=self.box.to_numpy().copy(),
            topology=self._topology.copy()
        )

    # Describe
    def describe(self):
        # print('protein sequence: %s' % self.topology.residues)
        print('# structures: %s' % self.n_structures)
        print('# atoms: %s' % self.n_atoms)
        print('# dimensions: %s' % self.n_dim)

    # Exclude
    def exclude(self, **kwargs):
        """
        Convenience function that passes to :func:`select`.

        Returns
        -------
        Trajectory
        """

        return self.select(exclude=True, **kwargs)

    def explode(self, n):
        # Explode
        xyz = pd.concat([self.xyz] * n)
        xyz.index = pd.MultiIndex.from_product([np.arange(n), self.xyz.index.get_level_values(1)],
                                               names=['structure_id', 'atom_id'])

        return Trajectory(data=xyz)

    # Get atoms
    def get_atoms(self, index):
        """
        Get specific atom indices from the Trajectory.

        Parameters
        ----------
        index : array-like
            List of atom indices.

        Returns
        -------
        pandas.DataFrame
            Trajectory information for atom indices.
        """

        # xyz will be a view if index is an int, otherwise it will be a copy...
        # TODO how will this affect behavior?
        # return self._xyz[:, index, :]
        mask = np.in1d(self._data.index.get_level_values('atom_id'), index)
        return self._data[mask].copy()  # need to do a copy here, because otherwise the view creates downstream warnings

    # Get column
    def get_column(self, column):
        return self._data[column]

    # Get structure
    def get_structure(self, index):
        """
        Get specific structure indices from the Trajectory.

        Parameters
        ----------
        index : array-like
            List of structure indices.

        Returns
        -------
        pandas.DataFrame
            Trajectory information for structure indices.
        """

        mask = np.in1d(self._data.index.get_level_values('structure_id'), index)
        return Trajectory(self._data[mask], topology=self._topology)

    def keys(self):
        return self._topology.keys()

    def min(self):
        return self.coordinates.pivot_table(index='structure_id', values=['x', 'y', 'z'], aggfunc='min')

    def max(self):
        return self.coordinates.pivot_table(index='structure_id', values=['x', 'y', 'z'], aggfunc='max')

    # Move
    @set_doc(move)
    def move(self, by=None, to=None, inplace=True):
        return move(self, by, to, inplace)

    # Query
    def query(self, expr, only_index=False):
        """
        Query the set `Topology` and return matching indices

        Parameters
        ----------
        expr : str
            Selection text. See :ref:`pandas.DataFrame.query`
        only_index : bool
            Should the entire Trajectory be returned, or only the array of indices?

        Returns
        -------
        numpy.ndarray or Trajectory
            Array of indices or Trajectory containing only expr.
        """

        # Query the topology to get pertinent data
        topology = self.topology.query(expr)

        # Extract indices
        # noinspection PyProtectedMember
        index = topology._data.index.to_numpy()

        #
        if only_index:
            result = index

        else:
            # TODO does this work for when multiple structures should be returned?
            result = Trajectory(self.get_atoms(index), topology=topology)  # noqa

        # Return result
        return result

    # Select
    # TODO compare the efficiency of this vs query
    def select(self, exclude=False, **kwargs):
        """

        Parameters
        ----------


        Returns
        -------

        """

        # Save snapshot of topology data
        # noinspection DuplicatedCode,DuplicatedCode
        data = self._topology.to_frame()

        # Continuously slice the topology data
        for key in kwargs:
            item = kwargs[key]
            if not isinstance(item, ArrayLike):
                item = [item]
            mask = data[key].isin(item)
            if exclude:
                mask = ~mask
            data = data[mask]

        # Extract indices and create a new topology
        # Remember: Topology index is atom_id
        # TODO should I inforce that Topology.index = Topology.atom_id??
        index = data['atom_id'].to_numpy()  # used to be data.index, but this breaks on multiple selections
        topology = Topology(data)
        assert len(index) == len(topology)  # noqa

        # Create new Trajectory, log, and return
        # return Trajectory(self.get_atoms(index).reshape(self.n_structures, len(index), self.n_dim), topology=topology)
        trajectory = Trajectory(data=self.get_atoms(index), topology=topology)  # noqa
        trajectory._configuration = self._configuration
        trajectory._parent = self if self._parent is None else self._parent  # keep track of selection origin
        logger.info(f'selected {trajectory.n_atoms} of {self.n_atoms} atoms (new Trajectory: '
                    f'{trajectory.designator}; exclude mode={exclude};'
                    f' {kwargs})')
        assert len(index) == trajectory.n_atoms  # noqa
        return trajectory

    # Move selection to periodic image
    def to_image(self, ix, iy, iz, weights=None, inplace=False):
        """
        Move selection to image.

        Parameters
        ----------
        ix
        iy
        iz
        weights : numpy.ndarray
            (Optional) Weights for atoms in the trajectory. Follows the definition from :ref:`numpy.average`.
        inplace : bool
            Make the change in place? (Default: False)

        Returns
        -------

        """

        if weights:
            raise AttributeError

        # Wrap to unit cell
        xyz = self.wrap_center(inplace=False)

        # Get box
        box = self.box.rename(columns={'bx': 'x', 'by': 'y', 'bz': 'z'})

        # Then, move to image
        xyz = xyz + np.array([ix, iy, iz]) * box

        # Do in place or return a copy TODO change this into a set_coordinates command
        trajectory = self if inplace else self.copy()
        trajectory.coordinates = xyz
        logger.info(f'moved {trajectory.designator} to image {(ix, iy, iz)}')
        if not inplace:
            return trajectory

    # Recenter the Trajectory at the origin
    # TODO need to decide if this should be moved to molecular.transform
    def to_origin(self, weights=None, inplace=False):
        """
        Recenter all structures in the Trajectory around the origin.

        Parameters
        ----------
        weights : numpy.ndarray
            (Optional) Weights for atoms in the trajectory. Follows the definition from :ref:`numpy.average`.
        inplace : bool
            Make the change in place? (Default: False)
        """

        # Center xyz coordinates
        coord = self.coordinates - self.center(weights)

        # Center Trajectory in place or return a copy
        trajectory = self if inplace else self.copy()
        trajectory.coordinates = coord
        logger.info(f'centered {trajectory.designator} at origin')
        if not inplace:
            return trajectory

    # Convert to pandas DataFrame
    def to_frame(self):
        """
        Convert `Trajectory` to pandas.DataFrame instance

        Returns
        -------
        pandas.DataFrame
            Trajectory represented as pandas DataFrame
        """

        # Return
        return self._data

    # Set structure IDs to a new range
    def set_structure_ids(self, structure_ids):
        """
        Set structure IDs in Trajectory.

        Parameters
        ----------
        structure_ids : list
        """

        # Set the values
        self._data.index = self._data.index.set_levels(structure_ids, level=0)  # noqa
        self._configuration.index = structure_ids

        # Make sure the index still has the correct name
        self._data.index.set_names(['structure_id', 'atom_id'], inplace=True)
        self._configuration.index.name = 'structure_id'

    # Show
    def show(self):
        pass

    # Save as PDB
    def to_pdb(self, fname):
        """
        Convert `Trajectory` to PDB.

        Parameters
        ----------
        fname : str
            Name of PDB to write.
        """

        # Convert topology and trajectory to DataFrame
        topology = self.topology.to_frame().reset_index()
        trajectory = self.to_frame().reset_index()

        # Merge trajectory and topology
        data = trajectory.merge(topology.reset_index(), how='inner', on='atom_id').set_index('atom_id').reset_index()

        # Sort by structure_id and then atom_id
        data = data.sort_values(['structure_id', 'atom_id'])

        # Add ATOM record
        data['record'] = 'ATOM'

        # TODO this should also be done for residue id probably
        # Change base-0 to base-1 for atom_id
        # if data['atom_id'].min() == 0:
        #     data['atom_id'] += 1

        # If we have more than 100000 atoms, we need to change atom_id
        # TODO this sort of sucks
        # https://github.com/MDAnalysis/mdanalysis/issues/1897
        fmt = '%-6s%5i %4s%4s%2s%4i%12.3f%8.3f%8.3f%6.2f%6.2f%9s%2s'
        if data['atom_id'].max() >= 99999:
            fmt = '%-6s%5s %4s%4s%2s%4i%12.3f%8.3f%8.3f%6.2f%6.2f%9s%2s'
            data['atom_id'] = pd.Series(np.vectorize(hex)(data['atom_id'])).str[2:]
            if data['atom_id'].str.len().max() >= 6:
                raise ValueError('cannot hex atom_id')
        else:
            # increment by 1 so atom_id starts at 1 and not 0
            data['atom_id'] = data['atom_id'] + 1

        # Format atom names
        i = data['atom'].str.len() == 1  # noqa
        data.loc[i, 'atom'] = data.loc[i, 'atom'].str.pad(2, side='left').str.pad(4, side='right')

        i = data['atom'].str.len() == 2
        data.loc[i, 'atom'] = data.loc[i, 'atom'].str.pad(3, side='left').str.pad(4, side='right')

        # Open up the buffer
        with open(fname, 'w') as buffer:
            # Select every structure and write out
            # TODO please make this more efficient
            for structure in data['structure_id'].unique():
                is_structure = data['structure_id'] == structure

                # Select pertinent rows and columns
                structure = data.loc[is_structure, ['record', 'atom_id', 'atom', 'residue', 'chain', 'residue_id',
                                                    'x', 'y', 'z', 'alpha', 'beta', 'segment', 'element']].to_numpy()

                # Write out PDB file
                # TODO add box to header
                np.savetxt(
                    buffer,
                    structure,
                    fmt=fmt,
                    header='CRYST1    0.000    0.000    0.000  90.00  90.00  90.00 P 1           1',
                    footer='END\n',
                    comments=''
                )

    # Convert to MDAnalysis Universe
    # mdanalysis.org
    def to_universe(self):
        pass

    # Update
    def update(self, xyz=None):
        if xyz is not None:
            if self.xyz.shape != xyz.shape:
                raise AttributeError('must be same shape')
            self.xyz = xyz

    # Wrap
    def wrap_center(self, inplace=True):
        """
        Wrap the center of the Trajectory to the unit cell

        Parameters
        ----------
        inplace : bool

        Returns
        -------

        """

        com = self.center()
        box = self.box.rename(columns={'bx': 'x', 'by': 'y', 'bz': 'z'})

        xyz = self.coordinates - box * (com / box).round()

        if inplace:
            self.coordinates = xyz
        else:
            return xyz

    # Wrap all
    def wrap_all(self, inplace=True):
        """
        Wraps all atoms in the selection to the unit cell.

        Parameters
        ----------
        inplace : bool
            Should the wrap happen in place? (Default: True)

        Returns
        -------
        Trajectory
        """

        # Wrap all atoms to the unit cell
        xyz = self.coordinates
        box = self.box.rename(columns={'bx': 'x', 'by': 'y', 'bz': 'z'})
        xyz_wrapped = xyz - box * (xyz / box).round()

        # Do in place or return a copy
        trajectory = self if inplace else self.copy()
        trajectory.coordinates = xyz_wrapped
        logger.info(f'wrapped all atoms of {trajectory.designator} to unit cell')
        if not inplace:
            return trajectory

    # View
    # https://github.com/arose/nglview
    def view(self):
        pass


# Topology class
# TODO what about charge?
class Topology:
    """
    A `Topology` is an object that stores metadata for a `Trajectory`, such as atomic relationships

    "Atomic" can be used loosely here because it applies to proteins, proteins in water baths, proteins in lipid
    environments, ligands in water, etc.

    The `Topology` instance generally follows PDB format [1] in that in contains,
      - atom_id
      - atom
      - residue
      - chain
      - residue_id
      - alpha
      - beta
      - segment
      - element
    """

    # Required columns
    columns = [
        'atom_id',
        'atom',
        'atom_type',
        'residue',
        'chain',  # only available if read from PDB
        'residue_id',
        'alpha',  # only available if read from PDB
        'beta',  # only available if read from PDB
        'segment',
        'element',  # only available if read from PDB
    ]

    @classmethod
    def from_universe(cls, universe):
        return from_universe(universe, return_topology=True, return_trajectory=False)

    # Initialize class instance
    def __init__(self, data=None):
        """
        Initialize class instance
        """

        # If data is set, check that it's a DataFrame
        if data is not None and not isinstance(data, pd.DataFrame):
            raise AttributeError('data must be pandas.DataFrame')

        # Add index to data
        if isinstance(data, pd.DataFrame):
            data['index'] = (data['atom_id'].rank() - 1).astype(int)
            data = data.set_index('index')

        # TODO can this be protected?
        # Save data or create blank DataFrame
        self._data = data if data is not None else pd.DataFrame(columns=self.columns)

    # Add atom to topology
    def __add__(self, other):
        pass

    # Get attribute
    def __getattr__(self, item):
        if item not in self._data:
            raise AttributeError('%s not in Structure' % item)

        return self._data[item].to_numpy()

    # Get atom from Structure by index
    def __getitem__(self, item):
        """


        Parameters
        ----------
        item

        Returns
        -------

        """

        # First, try to see if we can ping the DataFrame
        if isinstance(item, str):
            return self._data[item].to_numpy()

        # Make sure that item a valid atom_id
        if item not in self._data['atom_id']:  # noqa
            raise AttributeError('%s not a valid atom_id' % item)

        # Create copy of the row
        data = self._data[self._data['atom_id'] == item].copy()

        # Sanity
        if len(data) != 1:
            raise AttributeError('multiple atom_id %s found in Structure' % item)

        # Return (as Series)
        return data.iloc[0, :]

    # Length of structure (returns number of atoms)
    def __len__(self):
        """
        Number of atoms in the Structure instance.

        Returns
        -------
        int
            Number of atoms in the Structure instance.
        """

        return self.n_atoms

    # Trajectory designator
    @property
    def designator(self):
        return f'Topology:{self.hex_id}'

    # Hex ID
    @property
    def hex_id(self):
        return hex(id(self))

    # Add new atoms f
    def add_atoms(self, **kwargs):
        pass

    # Apply
    def apply(self, function):
        return function(self)

    # TODO is this too specific? Evaluate if Structure should be more generalized
    # Compute secondary structure
    # def compute_secondary_structure(self, recompute=False, executable='stride'):
    #     # If secondary_structure is not in the Structure, or recompute is true, compute
    #     if 'secondary_structure' not in self._data or recompute:
    #         # TODO in lieu of directly porting stride to Python, how can this be sped up?
    #         # Write out PDB of structure to glovebox
    #         gb = GloveBox('molecular-stride', persist=True)
    #         temp_pdb = os.path.join(gb.path, 'temp.pdb')
    #         self.to_pdb(temp_pdb)
    #
    #         # Run STRIDE
    #         secondary_structure = stride(temp_pdb, executable=executable)[['residue_id', 'secondary_structure']]
    #
    #         # Clean glovebox
    #         gb.clean()
    #
    #         # Drop secondary_structure if it's already in self._data
    #         self._data = self._data.drop(columns='secondary_structure', errors='ignore')
    #
    #         # Merge STRIDE results with Structure to store
    #         self._data = self._data.merge(secondary_structure, how='inner', on='residue_id', validate='m:1')
    #
    #     # Return
    #     return self._data[['residue_id', 'secondary_structure']].drop_duplicates()

    # Copy
    def copy(self):
        """
        Create a copy of `Topology` instance.

        Returns
        -------
        molecular.Topology
            Copy of instance.
        """

        return Topology(self._data.copy())

    # Infer elements from the Topology
    def infer_elements(self, overwrite=False, inplace=True):
        if not inplace:
            raise NotImplementedError

        # "Infer" from first letter of atom name; this will fail for two letter elements...
        # FIXME for 2 letter elements
        # TODO add warning for the fixme?
        elements = self._data['atom'].str.strip().str[0]
        logger.info(f'inferring missing elements from {self.designator} (overwrite = {overwrite})')
        if overwrite or 'element' not in self._data.columns:
            self._data['element'] = elements
        else:
            mask = self._data['element'].isnull()
            self._data.loc[mask, 'element'] = elements[mask]

    def keys(self):
        return self._data.columns.to_numpy()

    # Number of atoms
    @property
    def n_atoms(self):
        """
        Get the number of atoms in the `Topology`.

        Returns
        -------
        int
            Number of atoms
        """

        return len(self._data)

    @property
    def n_residues(self):
        return len(np.unique(self._data['residue_id']))

    # Pivot
    def pivot(self, index, columns=None, values=None, aggfunc='mean'):
        return pivot(self._data, index=index, columns=columns, values=values, aggfunc=aggfunc)

    # TODO should this return a Structure or Trajectory? TBD
    # TODO what if there were an intermediate object, like a Query class that knew how to handle subindexing / setting
    # Query
    def query(self, text):
        # TODO make this so it's customizable
        # Run macros
        text = text.lower().replace('peptide', 'residue in ["ALA", "ARG", "ASN", "ASP", "CYS", "GLN", "GLU", "GLY", ' +
                                    '"HIS", "HSD", "ILE", "LEU", "LYS", "MET", "PHE", "PRO", ' +
                                    '"SER", "THR", "TRP", "TYR", "VAL"]')

        # Get indices and parsed data
        indices = self._data.query(text).index.to_numpy()
        data = self._data.loc[self._data.index.isin(indices), :].copy()

        # Return
        return Topology(data)

    # TODO extend this to another trajectory
    # Compute RMSD
    def rmsd(self, structure):
        pass

    # Get residues
    @property
    def residues(self):
        """
        Get the residues of the Topology.

        Returns
        -------
        numpy.ndarray
            Array of residue names.
        """

        data = self._data[['residue_id', 'residue']].drop_duplicates()
        return data['residue'].to_numpy()

    # Write to CSV
    def to_csv(self, path, topology=False):
        pass

    # Convert to pandas DataFrame
    def to_frame(self):
        """
        Convert to pandas.DataFrame

        Returns
        -------
        pandas.DataFrame
            `Topology` as pandas.DataFrame
        """

        return self._data.copy()

    # Convert to trajectory
    def to_trajectory(self, structure_id=None):
        # Create and/or update structure_id if necessary
        if 'structure_id' not in self._data:
            structure_id = 0
        if structure_id is not None:
            self._data['structure_id'] = structure_id

        # Return Trajectory
        return Trajectory(self._data)

    # Get coordinates
    @property
    def xyz(self):
        """

        Returns
        -------

        """

        return self._data[['x', 'y', 'z']].to_numpy()


# Create Trajectory + Topology from universe
# TODO I don't know if this belongs here or in a different module. It definitely doesn't belong in IO.
# TODO should this always return both topology + trajectory?
def from_universe(universe, return_topology=True, return_trajectory=True):
    # Create Topology
    topology = None
    if hasattr(universe.atoms, 'names'):
        df = pd.DataFrame({
            'atom_id': universe.atoms.indices,
            'atom': universe.atoms.names,
            'atom_type': universe.atoms.types,
            'residue': universe.atoms.resnames,
            'residue_id': universe.atoms.resids,
            'segment': universe.atoms.segids,
        })

        topology = Topology(df)
        topology.infer_elements(inplace=True)

    # Create Trajectory
    trajectory = None
    if hasattr(universe, 'trajectory'):
        # Transfer to memory
        universe.transfer_to_memory()

        # Extract out box and xyz
        box = universe.trajectory.dimensions_array[:, :3]
        xyz = universe.trajectory.coordinate_array

        trajectory = Trajectory(coordinates=xyz, box=box)
        if topology is not None:
            trajectory._topology = topology  # TODO make this so you're not setting a private attribute

    # What to return?
    if return_topology and return_trajectory:
        return topology, trajectory
    elif return_topology:
        return topology
    elif return_trajectory:
        return trajectory


def _dummy_trajectory_data(n_structures, n_atoms, n_dim):
    # Sanity
    if not isinstance(n_structures, int) or not isinstance(n_atoms, int) or not isinstance(n_dim, int):
        raise AttributeError(f'n_structures = {n_structures}; n_atoms = {n_atoms}; n_dim = {n_dim}')
    if n_dim != 3:
        raise AttributeError('can only process 3D information for now')

    # Turn record counts into arrays of indices
    structure_ids = np.arange(n_structures)
    atom_ids = np.arange(n_atoms)

    # Return
    return pd.DataFrame(index=pd.MultiIndex.from_product([structure_ids, atom_ids], names=['structure_id', 'atom_id']))


if __name__ == '__main__':
    import molecular as mol

    trj = mol.read_pdb('../tests/samples/trajectory.pdb')




#cython: language_level=3

# see http://www.ks.uiuc.edu/Research/vmd/plugins/doxygen/dcdplugin_8c-source.html

import numpy as np
from scipy.io import FortranFile

cimport numpy as np

def _read_dcd(fname):
    return _get_box_and_coordinates(fname)

cdef tuple _get_box_and_coordinates(fname):
    # Declarations
    cdef np.ndarray[np.int32_t, ndim=1] iarr
    cdef np.int32_t n_str, n_atoms, i
    cdef np.ndarray[np.float64_t, ndim=2] box
    cdef np.ndarray[np.float32_t, ndim=3] xyz
    cdef np.ndarray[np.float64_t, ndim=1] r8arr

    # Open FortranFile buffer
    buffer = FortranFile(fname, 'r')

    # Header
    iarr = buffer.read_record('i')
    if iarr[0] != 1146244931 or iarr[9] != 0:  # iarr[0] == int.from_bytes(bytes(b'CORD'), 'little')
        # TODO flip byte order
        raise IOError('cannot parse DCD file')
    n_str = iarr[1]

    # Title
    iarr = buffer.read_record('i')
    if iarr[0] != 2:
        raise IOError('cannot parse DCD file')

    # Atoms
    iarr = buffer.read_record('i')
    n_atoms = iarr[0]

    # Box and coordinate information
    box = np.zeros((n_str, 3), dtype=np.float64)
    xyz = np.zeros((n_str, n_atoms, 3), dtype=np.float32)
    for i in range(n_str):
        r8arr = buffer.read_record('f8')
        box[i, :] = r8arr[[0, 2, 5]]

        xyz[i, :, 0] = buffer.read_record('f')
        xyz[i, :, 1] = buffer.read_record('f')
        xyz[i, :, 2] = buffer.read_record('f')

    # Close out buffer
    buffer.close()

    # Return
    return box, xyz




import numpy as np
import pandas as pd


class Residue:
    """

    """

    # This should just be a dataframe
    def __init__(self, name, charge):
        self._name = name
        self._charge = charge
        self._atoms = pd.DataFrame(columns=['group', 'name', 'kind', 'charge', 'comment'])
        self._bonds = pd.DataFrame(columns=['name0', 'name1'])
        self._impropers = pd.DataFrame(columns=['name0', 'name1', 'name2', 'name3'])

    @property
    def atoms(self):
        return self._atoms

    @property
    def bonds(self):
        return self._bonds

    @property
    def impropers(self):
        return self._impropers

    def add_atom(self, group, name, kind, charge, comment=''):
        self._atoms.loc[len(self._atoms)] = {
            'group': group,
            'name': name,
            'kind': kind,
            'charge': charge,
            'comment': comment
        }

    def add_bond(self, name0, name1):
        self._bonds.loc[len(self._bonds)] = {
            'name0': name0,
            'name1': name1
        }

    def add_improper(self, name0, name1, name2, name3):
        self._impropers.loc[len(self._impropers)] = {
            'name0': name0,
            'name1': name1,
            'name2': name2,
            'name3': name3
        }

    def compare_atoms(self, other):
        self_atoms = self.atoms.drop(columns='comment')
        other_atoms = other.atoms.drop(columns='comment')
        return self_atoms.merge(other_atoms, how='outer', indicator=True)

    def compare_bonds(self, other):
        bonds = self.bonds.merge(other.bonds, how='outer', indicator=True)
        return bonds

    def compare_impropers(self, other):
        return self.impropers.merge(other.impropers, how='outer', indicator=True)

    def copy(self):
        copy = Residue(name=self._name, charge=self._charge)
        copy._atoms = self._atoms.copy()
        copy._bonds = self._bonds.copy()
        copy._impropers = self._impropers.copy()
        return copy

    def find_bonds(self, names):
        bonds = self.bonds[np.max(np.isin(self._bonds, names), axis=1)]
        unique_bonds = np.unique(bonds.to_numpy().ravel())
        mask = ~np.isin(unique_bonds, names)
        return unique_bonds[mask]

    # Rename atom names in Residue
    def rename(self, old_to_new):
        """
        Rename atom names in this Residue. This includes ATOM information and BOND/IMPR information.

        Parameters
        ----------
        old_to_new : dict
            Old name is the key, new name is the value.
        """

        # Loop over all key/value pairs provided
        for old_name, new_name in old_to_new.items():
            # First check that new_name isn't already in the molecule
            if new_name in self.atoms['name']:
                raise AttributeError(f'cannot rename {old_name} to {new_name} because it already exists')

            # Make the change with atoms
            self.atoms.loc[self.atoms['name'] == old_name, 'name'] = new_name

            # Make the change with bonds
            self.bonds.loc[self.bonds['name0'] == old_name, 'name0'] = new_name
            self.bonds.loc[self.bonds['name1'] == old_name, 'name1'] = new_name

            # Make the change with impropers
            self.impropers.loc[self.impropers['name0'] == old_name, 'name0'] = new_name
            self.impropers.loc[self.impropers['name1'] == old_name, 'name1'] = new_name
            self.impropers.loc[self.impropers['name2'] == old_name, 'name2'] = new_name
            self.impropers.loc[self.impropers['name3'] == old_name, 'name3'] = new_name

    # Sort bonds
    def sort_bonds(self):
        # For convenience
        bonds = self.bonds

        # First, correct "order of order" bonds. Note that hydrogens are not sorted this way
        out_of_order = (bonds['name0'].str[0] == 'H') | \
                       ((bonds['name0'] > bonds['name1']) & (bonds['name1'].str[0] != 'H'))
        bonds['temp'] = bonds['name0']
        bonds.loc[out_of_order, 'name0'] = bonds.loc[out_of_order, 'name1']
        bonds.loc[out_of_order, 'name1'] = bonds.loc[out_of_order, 'temp']

        # Second, sort within DataFrame
        bonds['has_hydrogen'] = bonds['name1'].str[0] == 'H'
        bonds.sort_values(['has_hydrogen', 'name0', 'name1'], inplace=True)

        # Drop temp columns and reset index
        bonds.drop(columns=['temp', 'has_hydrogen'], inplace=True)
        bonds.reset_index(drop=True, inplace=True)


class Parameters:
    def __init__(self):
        self._bonds = pd.DataFrame(columns=['kind0', 'kind1', 'u', 'r', 'comment'])
        self._angles = pd.DataFrame(columns=['kind0', 'kind1', 'kind2', 'u', 'r', 'comment'])
        self._dihedrals = pd.DataFrame(columns=['kind0', 'kind1', 'kind2', 'kind3', 'u', 'n', 'r', 'comment'])

    @property
    def angles(self):
        return self._angles

    @property
    def bonds(self):
        return self._bonds

    @property
    def dihedrals(self):
        return self._dihedrals

    def add_angle(self, kind0, kind1, kind2, u, r, comment=''):
        self._angles.loc[len(self._angles)] = {
            'kind0': kind0,
            'kind1': kind1,
            'kind2': kind2,
            'u': u,
            'r': r,
            'comment': comment
        }

    def add_bond(self, kind0, kind1, u, r, comment=''):
        self._bonds.loc[len(self._bonds)] = {
            'kind0': kind0,
            'kind1': kind1,
            'u': u,
            'r': r,
            'comment': comment
        }

    def add_dihedral(self, kind0, kind1, kind2, kind3, u, n, r, comment=''):
        self._dihedrals.loc[len(self._dihedrals)] = {
            'kind0': kind0,
            'kind1': kind1,
            'kind2': kind2,
            'kind3': kind3,
            'u': u,
            'n': n,
            'r': r,
            'comment': comment
        }

    def copy(self):
        params = Parameters()
        params._bonds = self._bonds.copy()
        params._angles = self._angles.copy()
        params._dihedrals = self._dihedrals.copy()
        return params

    def merge(self, other):
        # Create new parameter set
        params = Parameters()

        # Bonds
        params._bonds = pd.concat([self.bonds, other.bonds], ignore_index=True).drop_duplicates()  # noqa
        params._bonds['has_hydrogen0'] = params._bonds['kind0'].str[0] == 'H'
        params._bonds['has_hydrogen1'] = params._bonds['kind1'].str[0] == 'H'
        params._bonds.sort_values(['has_hydrogen0', 'kind0', 'has_hydrogen1', 'kind1'], inplace=True)
        params._bonds.drop(columns=['has_hydrogen0', 'has_hydrogen1'], inplace=True)
        params._bonds.reset_index(drop=True, inplace=True)  # noqa

        # Angles
        params._angles = pd.concat([self.angles, other.angles], ignore_index=True).drop_duplicates()
        params._angles['has_hydrogen0'] = params._angles['kind0'].str[0] == 'H'
        params._angles['has_hydrogen1'] = params._angles['kind1'].str[0] == 'H'
        params._angles['has_hydrogen2'] = params._angles['kind2'].str[0] == 'H'
        params._angles.sort_values(['has_hydrogen0', 'kind0', 'has_hydrogen1', 'kind1', 'has_hydrogen2', 'kind2'],
                                   inplace=True)
        params._angles.drop(columns=['has_hydrogen0', 'has_hydrogen1', 'has_hydrogen2'], inplace=True)
        params._angles.reset_index(drop=True, inplace=True)  # noqa

        # Dihedrals
        params._dihedrals = pd.concat([self.dihedrals, other.dihedrals], ignore_index=True).drop_duplicates()
        params._dihedrals['has_hydrogen0'] = params._dihedrals['kind0'].str[0] == 'H'
        params._dihedrals['has_hydrogen1'] = params._dihedrals['kind1'].str[0] == 'H'
        params._dihedrals['has_hydrogen2'] = params._dihedrals['kind2'].str[0] == 'H'
        params._dihedrals['has_hydrogen3'] = params._dihedrals['kind3'].str[0] == 'H'
        params._dihedrals.sort_values \
            (['has_hydrogen0', 'kind0', 'has_hydrogen1', 'kind1', 'has_hydrogen2', 'kind2', 'has_hydrogen3', 'kind3',
              'n'], inplace=True)
        params._dihedrals.drop(columns=['has_hydrogen0', 'has_hydrogen1', 'has_hydrogen2', 'has_hydrogen3'],
                               inplace=True)
        params._dihedrals.reset_index(drop=True, inplace=True)

        # Return
        return params


class Stream:
    def __init__(self, fname=None):
        self._fname = fname
        self._header_rtf = ''
        self._header_prm = ''
        self._residue = None
        self._params = Parameters()

    @property
    def params(self):
        return self._params

    @property
    def residue(self):
        return self._residue

    # Create a copy of Stream object
    def copy(self):
        """
        Create a copy of the Stream object.
        """

        copy = Stream(self._fname)
        copy._header_rtf = self._header_rtf
        copy._header_prm = self._header_prm
        copy._residue = self._residue.copy()
        copy._params = self._params.copy()
        return copy

    def compare(self, other, errors='raise'):
        return self.equals(other, errors=errors)

    def equals(self, other, ignore_order=False, errors='raise'):
        # Are all atoms accounted for?
        diff_atoms = self.residue.compare_atoms(other.residue).query('_merge != "both"')
        if len(diff_atoms) != 0:
            if 'raise' in errors:
                raise AttributeError(diff_atoms)
            return False

        # Is the order of atoms the same?
        # noinspection PyUnresolvedReferences
        if not ignore_order and (self.residue.atoms['name'].to_numpy() != other.residue.atoms['name'].to_numpy()).any():
            return False

        # Are all bonds accounted for?
        self.residue.sort_bonds()
        other.residue.sort_bonds()
        diff_bonds = self.residue.compare_bonds(other.residue).query('_merge != "both"')
        if len(diff_bonds) != 0:
            if 'raise' in errors:
                raise AttributeError(diff_bonds)
            return False

        # Are all impropers accounted for?
        diff_imprs = self.residue.compare_impropers(other.residue).query('_merge != "both"')
        if len(diff_imprs) != 0:
            if 'raise' in errors:
                raise AttributeError(diff_imprs)
            return False

        # Are all bond parameters accounted for?
        self_bonds = self.params.bonds.drop(columns='comment')
        other_bonds = other.params.bonds.drop(columns='comment')
        diff_bonds = self_bonds.merge(other_bonds, how='outer', indicator=True).query('_merge != "both"')
        if len(diff_bonds) != 0:
            if 'raise' in errors:
                raise AttributeError(diff_bonds)
            return False

        # Are all angle parameters accounted for?
        self_angles = self.params.angles.drop(columns='comment')
        other_angles = other.params.angles.drop(columns='comment')
        diff_angles = self_angles.merge(other_angles, how='outer', indicator=True).query('_merge != "both"')
        if len(diff_angles) != 0:
            if 'raise' in errors:
                raise AttributeError(diff_angles)
            return False

        # Are all dihedral parameters accounted for?
        self_dihedrals = self.params.dihedrals.drop(columns='comment')
        other_dihedrals = other.params.dihedrals.drop(columns='comment')
        diff_dihedrals = self_dihedrals.merge(other_dihedrals, how='outer', indicator=True).query('_merge != "both"')
        if len(diff_dihedrals) != 0:
            if 'raise' in errors:
                raise AttributeError(diff_dihedrals)
            return False

        # Otherwise, return True
        return True

    def to_stream(self, fname):
        # Save
        with open(fname, 'w') as buf:
            buf.write(self._header_rtf)
            buf.write('RESI %-4s          %6s\n' % (self.residue._name, self.residue._charge))
            for group in self.residue.atoms['group'].unique():
                buf.write('GROUP\n')
                atoms = self.residue.atoms.query('group == %s' % group).copy()
                atoms['label'] = 'ATOM'
                # atoms[['label', 'name', 'kind', 'charge']].to_csv(buf, sep=' ', index=False, header=None)
                np.savetxt(buf, atoms[['label', 'name', 'kind', 'charge', 'comment']], fmt='%s %-6s %-6s %6s %s')
            buf.write('\n')
            bonds = self.residue.bonds.copy()
            bonds['label'] = 'BOND'
            # bonds[['label', 'name0', 'name1']].to_csv(buf, sep=' ', index=False, header=None)
            np.savetxt(buf, bonds[['label', 'name0', 'name1']], fmt='%s %-4s %-4s')
            buf.write('\n')
            imprs = self.residue.impropers.copy()
            imprs['label'] = 'IMPR'
            # imprs[['label', 'name0', 'name1', 'name2', 'name3']].to_csv(buf, sep=' ', index=False, header=None)
            np.savetxt(buf, imprs[['label', 'name0', 'name1', 'name2', 'name3']], fmt='%s %-6s %-6s %-6s %-6s')
            buf.write('\nEND\n')
            buf.write(self._header_prm)
            params = self.params.copy()
            buf.write('BONDS\n')
            # params.bonds[['kind0', 'kind1', 'u', 'r']].to_csv(buf, sep=' ', index=False, header=None)
            np.savetxt(buf, params.bonds[['kind0', 'kind1', 'u', 'r', 'comment']], fmt='%-6s %-6s %7s %10s %s')
            buf.write('\n')
            buf.write('ANGLES\n')
            # params.angles[['kind0', 'kind1', 'kind2', 'u', 'r']].to_csv(buf, sep=' ', index=False, header=None)
            np.savetxt(buf, params.angles[['kind0', 'kind1', 'kind2', 'u', 'r', 'comment']],
                       fmt='%-6s %-6s %-6s %7s %9s %s')
            buf.write('\n')
            buf.write('DIHEDRALS\n')
            # params.dihedrals[['kind0', 'kind1', 'kind2', 'kind3', 'u', 'n', 'r']].to_csv(buf, sep=' ', index=False, header=None)
            np.savetxt(buf, params.dihedrals[['kind0', 'kind1', 'kind2', 'kind3', 'u', 'n', 'r', 'comment']],
                       fmt='%-6s %-6s %-6s %-6s %10s %2s %8s %s')
            buf.write('\n')
            buf.write('IMPROPERS\n\nEND\nRETURN\n')


# Read stream file
# TODO move this to Cython
def read_stream(fname, sort=True):
    # Create object to store stream file data
    stream = Stream(fname)

    # Open stream file and read
    with open(fname, 'r') as buffer:
        # Start off in RTF header mode
        mode = 'header_rtf'

        # Loop over all lines
        for line in buffer.readlines():
            # Separate our lines into words surrounded by whitespace
            words = line.split()

            # Pick out the in-line comment, if it exists
            i = line.find('!')
            comment = line[i:].rstrip() if i >= 0 else ''

            # If the line is empty or we are not in header mode and find a comment, then skip
            if 'header' not in mode and (len(words) == 0 or words[0][0] == '!'):
                continue

            # Section 1 : header_rtf
            if mode == 'header_rtf' and (len(words) == 0 or words[0].lower() != 'resi'):
                stream._header_rtf += line

            # Section 2 : residue
            elif mode == 'header_rtf' and words[0].lower() == 'resi':
                mode = 'residue'
                group = -1
                stream._residue = Residue(words[1], words[2])

            elif mode == 'residue':
                if words[0].lower() == 'group':
                    group += 1
                elif words[0].lower() == 'atom':
                    stream._residue.add_atom(
                        group,
                        words[1],
                        words[2],
                        words[3],
                        # comment
                    )
                elif words[0].lower() == 'bond':
                    stream._residue.add_bond(
                        words[1],
                        words[2]
                    )
                elif words[0].lower() == 'impr':
                    stream._residue.add_improper(
                        words[1],
                        words[2],
                        words[3],
                        words[4]
                    )
                elif words[0].lower() == 'end':
                    mode = 'header_prm'
                else:
                    raise AttributeError('unexpected: %s' % line)

            # Section 3 : header_prm
            elif mode == 'header_prm' and (len(words) == 0 or words[0].lower() != 'bonds'):
                stream._header_prm += line

            # Section 4 : bonds
            elif mode == 'header_prm' and words[0].lower() == 'bonds':
                mode = 'bonds'

            elif mode == 'bonds' and words[0].lower() != 'angles':
                stream._params.add_bond(
                    words[0],
                    words[1],
                    words[2],
                    words[3],
                    comment
                )

            # Section 5 : angles
            elif mode == 'bonds' and words[0].lower() == 'angles':
                mode = 'angles'

            elif mode == 'angles' and words[0].lower() != 'dihedrals':
                stream._params.add_angle(
                    words[0],
                    words[1],
                    words[2],
                    words[3],
                    words[4],
                    comment
                )

            # Section 6 : dihedrals
            elif mode == 'angles' and words[0].lower() == 'dihedrals':
                mode = 'dihedrals'

            elif mode == 'dihedrals' and words[0].lower() != 'impropers':
                if len(words) > 7 and words[7][0] != '!':
                    raise AttributeError('unexpected: %s' % line)
                stream._params.add_dihedral(
                    words[0],
                    words[1],
                    words[2],
                    words[3],
                    words[4],
                    words[5],
                    words[6],
                    comment
                )

            # Section 7 : impropers
            elif mode == 'dihedrals' and words[0].lower() == 'impropers':
                mode = 'impropers'

            elif mode == 'impropers' and words[0].lower() not in ['end', 'return']:
                raise AttributeError('unexpected: %s' % line)

    # Sort?
    if sort:
        stream.residue.sort_bonds()

    # Return
    return stream


# Create DualStream object
# TODO name_map currently needs to be hand-mapped but in the future this could be automated
def create_dual_stream(wt, mt, name_map):
    """
    Create a dual topology Stream

    `name_map` is a dictionary that contains all atoms that change between WT and MT. If the atom name is the same in WT
    and MT, the expectation is that we will give the MT version a new name. If the atom is present in WT but not in MT,
    then it gets a value of None. If the atom is present in MT but not WT, we can keep its name.
    # First off, this is a list of all modified atoms.
    # Second, this tells us how to order modified atoms in the new dual stream and what to name them as.
    # If key and value are both values but different, this is what will be copied from WT to dual
    # If key and value are the same, this is an introduced atom in MT
    # If key is defined but value is None, then this is an atom that is removed in WT


    Parameters
    ----------
    wt : Stream
        Reference wild-type (wt) stream file
    mt : Stream
        Updated mutant (mt) stream file
    name_map : dict
        Dictionary that maps changed atom names in the current `wt` and `mt` stream files to the new combined file.

    Returns
    -------
    DualStream
    """

    # I'm not settled on this DualStream class, would it be better just to add new variables to Stream?
    return DualStream(wt, mt, name_map)


# Decompose Stream into wild-type and mutant
def decompose_dual_stream(ds, flags):
    """
    Decompose a Steam object into wild-type and mutant Stream objects.

    Parameters
    ----------
    ds : Stream
    flags : numpy.ndarray
        Follows NAMD-style where "0" represents common atom, "-1" represents wild-type atom, and "1" represent mutant
        atom.

    Returns
    -------
    Stream, Stream
        Wild-type and mutant Stream objects
    """

    # First, create copies of the Stream object fot wt and mt
    wt = ds.copy()
    mt = ds.copy()

    # Second, drop Residue info where flag False
    wt.residue._atoms = wt.residue.atoms[flags < 1].copy()
    mt.residue._atoms = mt.residue.atoms[flags > -1].copy()

    wt_atoms = wt.residue.atoms['name'].to_numpy()  # get atom names that are left
    mt_atoms = mt.residue.atoms['name'].to_numpy()

    wt.residue._bonds = wt.residue.bonds[
        (wt.residue.bonds['name0'].isin(wt_atoms)) & (wt.residue.bonds['name1'].isin(wt_atoms))]
    mt.residue._bonds = mt.residue.bonds[
        (mt.residue.bonds['name0'].isin(mt_atoms)) & (mt.residue.bonds['name1'].isin(mt_atoms))]

    wt.residue._impropers = wt.residue.impropers[
        (wt.residue.impropers['name0'].isin(wt_atoms)) & (wt.residue.impropers['name1'].isin(wt_atoms)) & (
            wt.residue.impropers['name2'].isin(wt_atoms)) & (wt.residue.impropers['name3'].isin(wt_atoms))]
    mt.residue._impropers = mt.residue.impropers[
        (mt.residue.impropers['name0'].isin(mt_atoms)) & (mt.residue.impropers['name1'].isin(mt_atoms)) & (
            mt.residue.impropers['name2'].isin(mt_atoms)) & (mt.residue.impropers['name3'].isin(mt_atoms))]

    # Third, drop Parameters where flag False
    wt_kinds = wt.residue.atoms['kind'].to_numpy()
    mt_kinds = mt.residue.atoms['kind'].to_numpy()

    wt.params._bonds = wt.params.bonds[
        (wt.params.bonds['kind0'].isin(wt_kinds)) & (wt.params.bonds['kind1'].isin(wt_kinds))]
    mt.params._bonds = mt.params.bonds[
        (mt.params.bonds['kind0'].isin(mt_kinds)) & (mt.params.bonds['kind1'].isin(mt_kinds))]

    wt.params._angles = wt.params.angles[
        (wt.params.angles['kind0'].isin(wt_kinds)) & (wt.params.angles['kind1'].isin(wt_kinds)) & (
            wt.params.angles['kind2'].isin(wt_kinds))]
    mt.params._angles = mt.params.angles[
        (mt.params.angles['kind0'].isin(mt_kinds)) & (mt.params.angles['kind1'].isin(mt_kinds)) & (
            mt.params.angles['kind2'].isin(mt_kinds))]

    wt.params._dihedrals = wt.params.dihedrals[
        (wt.params.dihedrals['kind0'].isin(wt_kinds)) & (wt.params.dihedrals['kind1'].isin(wt_kinds)) & (
            wt.params.dihedrals['kind2'].isin(wt_kinds)) & (wt.params.dihedrals['kind3'].isin(wt_kinds))]
    mt.params._dihedrals = mt.params.dihedrals[
        (mt.params.dihedrals['kind0'].isin(mt_kinds)) & (mt.params.dihedrals['kind1'].isin(mt_kinds)) & (
            mt.params.dihedrals['kind2'].isin(mt_kinds)) & (mt.params.dihedrals['kind3'].isin(mt_kinds))]

    # Return
    return wt, mt


# Subclass of Stream to hold dual topology streams
# TODO evaluate if this is worth it. Stream could also be used for this.
class DualStream(Stream):
    def __init__(self, wt, mt, name_map):
        # Create pointers to wt and mt Stream objects
        self._wt = wt
        self._mt = mt

        # Store name map
        self._name_map = name_map

        # Initialize this instance
        super().__init__()

        # Get mod atoms
        self._compute_mod_atoms()

        # Merge streams
        self._merge_streams()

    def _compute_mod_atoms(self):
        # For convenience
        wt = self._wt
        mt = self._mt
        name_map = self._name_map

        # Find atoms that were modified
        _mod_atoms = wt.residue.compare_atoms(mt.residue).query('_merge != "both"')['name'].unique()

        # Find hanging atoms attached to modified
        _mod_atoms_new = []
        for atom0 in _mod_atoms:
            for atom1 in wt.residue.find_bonds(atom0):
                if atom0 == atom1:
                    continue
                if len(wt.residue.find_bonds(atom1)) == 1:
                    _mod_atoms_new.append(atom1)
            for atom1 in mt.residue.find_bonds(atom0):
                if atom0 == atom1:
                    continue
                if len(mt.residue.find_bonds(atom1)) == 1:
                    _mod_atoms_new.append(atom1)
        mod_atoms = np.unique(np.hstack([_mod_atoms, np.array(_mod_atoms_new)]))

        # Mod atom sanity check; make sure that all names are accounted for
        for atom in mod_atoms:
            if atom not in name_map.keys():
                raise AttributeError('%s not found in names' % atom)
        for atom in name_map.keys():
            if atom not in mod_atoms:
                raise AttributeError('%s not found in mod_atoms' % atom)
        for atom in name_map.values():
            if atom in wt.residue.atoms['name']:
                raise AttributeError('%s in WT atoms already' % atom)

        # Save
        self._mod_atoms = mod_atoms

    #
    def _merge_streams(self):
        # For convenience
        wt = self._wt
        mt = self._mt
        name_map = self._name_map

        # Save header information
        self._header_rtf = wt._header_rtf
        self._header_prm = wt._header_prm

        # Build dual topology file
        residue = wt.residue.copy()
        for old_name, new_name in name_map.items():
            if new_name is None:
                continue
            atom = mt.residue.atoms.query('name == "%s"' % old_name)
            assert len(atom) == 1, old_name
            atom = atom.iloc[0]
            residue.add_atom(
                group=atom.group,
                name=new_name,
                kind=atom.kind,
                charge=atom.charge,
                comment=atom.comment
            )
            for _bond in mt.residue.find_bonds(old_name):
                bond = name_map.get(_bond, _bond)
                if (residue.find_bonds(new_name) == bond).any():  # noqa
                    continue
                residue.add_bond(
                    name0=new_name,
                    name1=bond
                )
            for _, _impr in mt.residue.impropers[np.max(np.isin(mt.residue.impropers, old_name), axis=1)].iterrows():
                impr = {}
                for key, value in _impr.items():
                    impr[key] = name_map.get(value, value)
                if (residue.impropers == impr).min(axis=1).sum() == 0:
                    residue.add_improper(
                        name0=impr['name0'],
                        name1=impr['name1'],
                        name2=impr['name2'],
                        name3=impr['name3']
                    )

        # Annotate dual topology
        wt_atoms = wt.residue.atoms['name'].to_numpy()
        mt_atoms = mt.residue.atoms['name'].to_numpy()
        for i, atom in enumerate(residue.atoms['name'].to_numpy()):
            in_wt = atom in wt_atoms
            in_mt = atom in mt_atoms
            lam = 0
            if (in_wt and not in_mt) or ((atom in name_map.keys()) and (atom not in name_map.values())):
                lam = -1
            elif (not in_wt and in_mt) or (atom in name_map.values()):
                lam = 1
            residue.atoms.loc[i, 'comment'] = residue.atoms.loc[i, 'comment'] + '! lam= %2s' % lam

        for old_name, new_name in name_map.items():
            mask = residue.atoms['name'] == new_name
            residue.atoms.loc[mask, 'comment'] = residue.atoms.loc[mask, 'comment'] + ', from %s (%s)' % (old_name,
                                                                                                          mt._fname)

        # Build dual parameters
        params = wt.params.merge(mt.params)

        # Save
        self._residue = residue
        self._params = params



import numpy as np
import pandas as pd


class Histogram:
    def __init__(self, data, weights=None):
        self._data = data

        # FIXME for non-ihist, the width will impact the mass function
        if weights is None:
            weights = 1.

        # self._probs = counts / np.sum(counts)  # pmf

    @property
    def counts(self):
        return self._data

    @property
    def pmf(self):
        """

        Returns
        -------

        """

        df = self._data / self._data.sum(axis=0)

        if len(df.columns) > 1:
            df.columns.name = 'pmf'
        else:
            df.rename(columns={'count': 'pmf'}, inplace=True)

        return df

    def plot(self):
        import uplot as u
        u.plot(self.to_frame())

    # FIXME when count = False and probability = False, this will break because there will be a pandas index with no
    #  other content.
    def to_frame(self, count=True, probability=False):
        # df = pd.DataFrame({'bin': self._bins})
        # if count:
        #     df['count'] = self._counts
        # if probability:
        #     df['probability'] = self._probs
        #
        # return df.set_index('bin')
        return self._data


def hist(a, bins):
    """
    Classical histogram, numpy style.

    Parameters
    ----------
    a
    bins

    Returns
    -------

    """

    pass


def ihist(a, width, right=False):
    """
    Create an interval histogram, i.e., given a known `width`, count all the occurrences of `a` in all bins :math:`[
    width, 2*width)`. By default, `right = False`. If true, then we look at :math:`(width, 2*width]` instead.

    Parameters
    ----------
    a : array
        Collection of values to turn into histogram.
    width : number
        The interval that should be binned.
    right : bool
        Left or right inclusive? If False, left inclusive. (Default: False)

    Returns
    -------
    Histogram
        Object that contains the histogram.
    """

    # Take `a` and get binned values
    func = np.floor if not right else np.ceil
    a_binned = func(a / width) * width

    # Create histogram
    if isinstance(a_binned, pd.DataFrame):
        a_melt = pd.melt(a_binned, value_name='bin', var_name='count')
        data = pd.crosstab(a_melt['bin'], a_melt['count'])
    else:
        bins, counts = np.unique(a_binned, return_counts=True)
        data = pd.DataFrame({'bin': bins, 'count': counts}).set_index('bin')

    # Return Histogram object
    return Histogram(data)


if __name__ == '__main__':
    a = np.random.rand(100000)
    print(ihist(a, 0.1).to_frame(probability=True, count=False))




import numpy as np

def angle(atom0, atom1, atom2, method='atan2', degrees=False):
    assert atom0.shape == atom1.shape == atom2.shape

    a = atom0.xyz.to_numpy() - atom1.xyz.to_numpy()
    b = atom2.xyz.to_numpy() - atom1.xyz.to_numpy()

    if method.lower() in 'atan2':
        results = np.arctan2(np.linalg.norm(np.cross(a, b, axis=1), axis=1), np.sum(a * b, axis=1))
    elif method.lower() in 'acos':
        results =  np.arccos(np.sum(a * b, axis=1) / (np.linalg.norm(a, axis=1) * np.linalg.norm(b, axis=1)))
    else:
        raise AttributeError(f'method {method} not understood')

    if degrees:
        results = np.rad2deg(results)

    return results